###### Standardized Phylogenetic Independent Contrasts Using 2D Geometric Morphometric Shape Data #####
# Travis Hagey

# Following Freckleton and Harvey, 2006, the node height test compares 
# standardized phylogenetic independent contrasts to the distance of each node above
# the root. Phylogenetic Independent Contrasts (PIC's) represent an estimate of the
# amount of morphological evolution that occurred between two sister taxa given the
# time they've been evolving independently, ie a local estimate of the rate of 
# morphological evolution. If the trait has been evolving similarly to Brownian 
# Motion, this rate should be constant throughout the tree. If the trait evolved 
# faster earlier in the clade's history, standardized contrasts should 
# be higher in nodes closer to the root. 

# Here we implement the above approach using generalized Procrustes aligned 
# 2D geometric morphometric shape data as our trait data. The approach is in essence replicating 
# traditional PIC but with high dimensional shape data. To estimate raw contrasts, 
# instead of finding the difference between focal tip values, we instead used
# Procrustes distances between tip shapes. Similarly, to estimate new condensed tip 
# values (ie shapes), we used the traditional PIC "weighted average" approach, but instead,
# estimating a mean shape, weighted by the two focal branch lengths. 

# The below functions assume the geomorph and phytools libraries are loaded

###### *** Included Functions *** #####

###### wt.mshape() - Calculate Weighted Mean Shape #####

wt.mshape <- function (shape.i, shape.j, weight.i, weight.j){
  
  ## Calculates the weighted average shape between two tip shapes, using the corresponding branch lengths as weights for the average 
    # The below calculations work if the input trait data (shape.i and shape.j) are univariate or 2D shapes
  
  ## Arguments 
  # shape.i and shape.j - the two tip values (or shapes) to be averaged (i.e. trait values at tips i and j)
    # if shape.i and shape.j are shapes, they should be 2D arrays (p x k) containing generalized Procrustes analysis-aligned specimen coordinates
  # weight.i and weight.j - the weights (i.e. branch lengths) that correspond to shape.i and shape.j
  
  ## Output 
  # shape.k - the newly estimated weighted average trait value (or shape) at the now condensed tip k. 

  shape.k <- ( ((1/weight.i)*shape.i) + ((1/weight.j)*shape.j) ) / ( (1/weight.i) + (1/weight.j) )
  return(shape.k)
  
}


###### std.cont() - Calculate Standardized Independent Contrast Using Procrustes Distances #####

std.cont <- function (shape.i, shape.j, weight.i, weight.j){
  
  ## Calculates the standardized independent contrast between two tips, using the Procrustes distance between shapes as the raw contrast
    # Note that the raw contrasts are divided by the square root of the summed branch lengths, which is actually the standard deviation, not the variance 
  
  ## Arguments 
  # shape.i and shape.j - the two shapes to be averaged (i.e. shapes at nodes i and j)
    # shape.i and shape.j should be 2D arrays (p x k) containing generalized Procrustes analysis-aligned specimen coordinates
  # weight.i and weight.j - the weights (i.e. branch lengths) that correspond to shape.i and shape.j
  
  # shape.i and shape.j can also be univaraite values for diagnostic purposes 
  
  ## Output 
  # S.ij - the standardized independent contrast between nodes i and j
  
  if(length(dim(shape.i))==2){
    cont.ij <- sqrt(sum((shape.i-shape.j)^2))  # Procrustes distance between two shapes (raw contrasts)
  }
  
  if(length(shape.i)==1){
    cont.ij <- abs(shape.i-shape.j)  # Univariate distance between two trait values (raw contrasts)
  }
  
  sdtcont.ij<- cont.ij / sqrt(weight.i + weight.j)
  return(sdtcont.ij)
  
}

##Diagnostic Code
  #std.cont(shape.i = dat.uni[1], shape.j = dat.uni[2], weight.i = 20, weight.j = 20)
  #std.cont(shape.i = dat.shape[,,1], shape.j = dat.shape[,,2], weight.i = 20, weight.j = 20)




###### new.branch.length() - Calculate Condensed Branch Length #####

new.branch.length <- function(length.i, length.j){
  ## Calculates the branch length to add to the existing branching leading to node k after the node is condensed
    # Note that this code does not add the existing branch length leading to node k. 
    # In the NH.Test function, instead of dropping both branches leading from k and adjusting the branch length leading to node k, 
    # we instead modified the length of one branch leading from node k to have the length that needs to be added,
    # and dropped the other branch, and reassigned the tip information of the remaining branch to be the new node k tip data
  
  ## Arguments 
  # length.i and length.j - the two condensed branch lengths that were connected to node k

  ## Output 
  # length.i.prime - the branch length to add to the branch leading to node k, after the two descending branches have been dropped
  
  length.i.prime<- (length.i*length.j) / (length.i+length.j) 
  return(length.i.prime)
}


##### PIC.shape() - Conducts PIC using Shape Data #####

PIC.shape <- function (phy, dat.shape){
  ## Calculates standardized phylogenetic independent contrasts using shape data and conducts a node height test
  
  ## Arguments
  # phy - a phylogeny with meaningful branch lengths. 
    #Tip names that do not match specimen names in dat.shape will be dropped

  # dat.shape - tip trait data
    # Specimen names in dat.shape must all correspond to tip names in phy
    # Specimen names not in phy will be removed from the dataset
    # dat.shape specimens do not need to be in the same order as the tips in phy

    # dat.shape can be a 3D array (p x k x n) containing Procrustes shape variables for a set of aligned specimens.
      # In this case dat.shape should  be the output coordinates data from the gpagen() function (i.e. dat.super$coords)
    # dat.shape can also be univariate data with assigned tip names for diagnostic purposes 

  
  ## Match and sort trait data and phy tips
    # Dropping tips not in data.shape, and dropping dat.shape individuals not in tree
  
  if(is.null(dim(dat.shape))) {message("dat.shape is univariate, not a 3D array (p x k x n) of x, y coordinates")}
  
  # Remove phy tips not in shape data
  
  if(length(dim(dat.shape))==3) { # if shape data
    message(paste("Following tips dropped from phylogeny:", phy$tip.label[!phy$tip.label %in% dimnames(dat.shape)[[3]]]))
    phy<-drop.tip(phy, phy$tip.label[!phy$tip.label %in% dimnames(dat.shape)[[3]]])
  }
  
  if(is.null(dim(dat.shape))) { # if univariate 
    message(paste("Following tips dropped from phylogeny:", phy$tip.label[!phy$tip.label %in% names(dat.shape)]))
    phy<-drop.tip(phy, phy$tip.label[!phy$tip.label %in% names(dat.shape)])
  }
  
  # Reorder shape data to only include tip data and be in same order as phy tips
  if(length(dim(dat.shape))==3) { # if shape data
    message(paste("Data not in phylogeny dropped:", dimnames(dat.shape)[[3]][!dimnames(dat.shape)[[3]] %in% phy$tip.label]))
    dat.shape<-dat.shape[,,phy$tip.label]
    } 
  
  if(is.null(dim(dat.shape))) { # if univariate 
    message(paste("Data not in phylogeny dropped:", names(dat.shape)[!names(dat.shape) %in% phy$tip.label]))
    dat.shape<-dat.shape[phy$tip.label]
    } 
  
  
  # Check shape data == phy tip names
  if(length(dim(dat.shape))==3) { 
    if(any(!dimnames(dat.shape)[[3]] == phy$tip.label)) {stop("Error matching tips and shape data")}
  }
  
  if(is.null(dim(dat.shape))){
    if(any(!names(dat.shape) == phy$tip.label)) {stop("Error matching tips and shape data")}
  }
  
  

  # Generate output object
  std.conts<-integer()
  
  
  # Work systematically through the phylogeny, calculating standard independent contrasts and condensing nodes  
  
  for(n in 1:phy$Nnode){
    
    # Find a node with only two total descendants (meaning its connected to two tips, and focus on the 1st in the list)
    current.node<-which(node.depth(phy, method=1)==2)[1] #grab one of the nodes in the list of penultimate nodes

    current.branches<-which(phy$edge[,1]==current.node) #which branches descend from the focal node
  
    current.length<-phy$edge.length[current.branches] #get the branch lengths for the two focal branches
  
    current.tips<-phy$edge[current.branches, 2] #get the two tips connected to the current node
  
  
    # Calculate Output Data
  
    output<-c(NA, NA)
    
    if(length(dim(dat.shape))==3) { 
      output[1]<-std.cont(shape.i = dat.shape[,,current.tips[1]], shape.j = dat.shape[,,current.tips[2]], weight.i = current.length[1], weight.j = current.length[2])
    }
    if(is.null(dim(dat.shape))){
      output[1]<-std.cont(shape.i = dat.shape[current.tips[1]], shape.j = dat.shape[current.tips[2]], weight.i = current.length[1], weight.j = current.length[2])
    }
    
    output[2]<-nodeheight(phy, current.node)
  
    std.conts<-rbind(std.conts, output)

    #when down to only two tips, stop
    if(Ntip(phy)==2){break}
    
    
    # Condense Branches
    #keep/modify 2nd branch, drop 1st branch
    
    phy$edge.length[current.branches[2]]<-new.branch.length(length.i = current.length[1], length.j = current.length[2])
    phy$tip.label[current.tips[2]]<-paste("Tip", (Ntip(phy)-1), sep="")
    phy<-drop.tip(phy, phy$tip.label[current.tips[1]])
    
    
    # Calculate weighted mean shape for new condensed node
    
    if(length(dim(dat.shape))==3){
      dat.shape[,,current.tips[2]]<-wt.mshape(shape.i = dat.shape[,,current.tips[1]], shape.j = dat.shape[,,current.tips[2]], weight.i = current.length[1], weight.j = current.length[2])
      dimnames(dat.shape)[[3]][current.tips[2]]<-paste("Tip", Ntip(phy), sep="") #revise new tip's name
      dat.shape<-dat.shape[,,-current.tips[1]] #drop removed shape
      # Check shape data == phy tip names
      if(any(!dimnames(dat.shape)[[3]] == phy$tip.label)) {stop("Error condensing nodes")}
    }
    
    if(is.null(dim(dat.shape))){
      dat.shape[current.tips[2]]<-wt.mshape(shape.i = dat.shape[current.tips[1]], shape.j = dat.shape[current.tips[2]], weight.i = current.length[1], weight.j = current.length[2])
      names(dat.shape)[current.tips[2]]<-paste("Tip", Ntip(phy), sep="") #revise new tip's name
      dat.shape<-dat.shape[-current.tips[1]] #drop removed tip data
      # Check shape data == phy tip names
      if(any(!names(dat.shape) == phy$tip.label)) {stop("Error condensing nodes")}
    }
   
  }
  
  colnames(std.conts)<-c("Std.IC", "NodeHeights")
  std.conts<-data.frame(std.conts)
  rownames(std.conts)<-NULL
  return(std.conts)
  
}

#####
##### Diagnostic/Example Code #####

if(FALSE) { 
  
  #Run the below code to generate an example PIC dataset
  
  attributes(plethodon) # example data modified from data(plethodon)

  Y.gpa<-gpagen(plethodon$land)    #GPA-alignment

  #Will focus on a single specimen from plethodon
  plotRefToTarget(Y.gpa$coords[,,1], Y.gpa$coords[,,1], method="vector", outline=plethodon$outline, mag=2.5)

  #copy specimen 1 four times into new array, named Specimen1, Specimen2, Specimen3, and Specimen4, Specimen5
  dat.pleth<-array(dim=c(12,2,5), dimnames = list(NULL, NULL, c("Specimen1", "Specimen2", "Specimen3", "Specimen4", "Specimen5")))
  dat.pleth[,,1:5]<-Y.gpa$coords[,,1]

  #Shift specimens 3 and 4 upwards to give them a known difference
  dat.pleth[,2,"Specimen3"]<-dat.pleth[,2,"Specimen3"]+sqrt(0.04/12)
  dat.pleth[,2,"Specimen4"]<-dat.pleth[,2,"Specimen4"]+sqrt(0.04/12)+sqrt(0.04/12)
  
  plotRefToTarget(dat.pleth[,,3], dat.pleth[,,4], method="vector", outline=plethodon$outline, mag=2.5)
  plotRefToTarget(dat.pleth[,,3], dat.pleth[,,1], method="vector", outline=plethodon$outline, mag=2.5)

  sqrt(sum((dat.pleth[,,1]-dat.pleth[,,4])^2))
  sqrt(sum((dat.pleth[,,3]-dat.pleth[,,1])^2))
  
  #with the above vertical shifts, specimen 1, 2, 5 are still identical
  #specimen 3 is has a Procrustes distance of 0.2 from specimens 1,2,5
  #specimen 4 is a further 0.2 from specimen 3, for a total distance of 0.4 from 
  #specimens 1,2,5
  
  # These simple values will help for easy tracking through the PIC process
  
  # univariate data
  dat.uni<-c(4.0, 4.0, 4.2, 4.4, 4.0)
  names(dat.uni)<-c("Specimen1", "Specimen2", "Specimen3", "Specimen4", "Specimen5")
  #univariate data follows similar pattern
  #specimens 1,2,5 are the same
  #specimen 3 is 0.2 larger 
  #specimen 4 is 0.2 larger still 

  
  #make phylogeny
  phy<-read.tree(text="(((Specimen1:20,Specimen2:20):10,(Specimen3:10,Specimen4:10):20):70,Specimen6:100);")
  plot(phy)
  edgelabels(phy$edge.length, bg="black", col="white", font=2)
  #this phylogney lacks species 5 from the data, and has species 6, for which there is no data
  
  
  to.export<-list(dat.pleth, dat.uni, phy)
  names(to.export)<-c("dat.pleth", "dat.uni", "phy")
  
  #set working directory to where you want to save
  save(to.export, file = "Demo_data.Rdata")

  
}


#load the above generated data
if(FALSE) {
  load("/Users/travishagey/Desktop/Hagey/Science/Projects/R codes/GM plots/PIC Shape Node Height Test/Demo_data.Rdata")
  phy<-to.export$phy
  dat<-to.export$dat.uni # should the trait data be univariate? 
  dat<-to.export$dat.pleth # should the trait data be shape? 
}


##Diagnostic wt.mshape() Code
if (FALSE){
  
  #if uni data
  wt.mshape(shape.i = dat[3], shape.j = dat[4], weight.i = 20, weight.j = 10)
  
  #if shape data
  shape.k<-wt.mshape(shape.i = dat[,,1], shape.j = dat[,,3], weight.i = 50, weight.j = 10)
  plot(dat[,,1], asp=1, cex=0.5, col="red")
  points(dat[,,3], cex=0.5, col="blue")
  points(shape.k, cex=0.5, col="green")
  
}


##Diagnostic std.cont() Code

if(FALSE){
  
  #if uni data
  std.cont(shape.i = dat[3], shape.j = dat[4], weight.i = 20, weight.j = 20)
  #should be 4.2-4.4 / sqrt(20+20)
  
  
  #if shape data
  std.cont(shape.i = dat[,,3], shape.j = dat[,,4], weight.i = 20, weight.j = 20)
  #should be 4.2-4.4 / sqrt(20+20)
  
}


##Diagnostic PIC.shape() Code
if(FALSE){
  
  #if uni data
  PIC.shape(phy = phy, dat.shape = dat)
  #should drop specimen 6 from tree
  #should drop specimen 5 from data
  
  #specimen 1 and 2 are the same, so they will have a standardized contrast of 0
  #their weighted average will still be the same
  
  #Specimen 3 and 4 have a difference (or distance) of 0.2, and divided by sqrt(10+10) = 0.0447
  #these two will have an average of 4.3 (or an weighted average shape that we know is 0.3 from specimne 1 and 2)
  
  #the new branch lengths will be 20 + (100/20) = 25 and 10 + (400/40) = 20
  
  #lastly 0.3 / sqrt(45) = again = 0.0447
  
  
}



#####


