#clear all current variables or functions from R memory and start (or restart) script/analyses
rm(list = ls(all = TRUE)) 

###### McCann Parallel toe pad shape comparisons ######

## Necessary files for this R script

  # McCann_Shape.tps - includes shape data from TPSdig
  # McCann_Slider.txt - three column file that indicates which two landmarks each semilandmark is located between
  # McCann_Pad_Categories.txt includes family and toe pad class assignments for each specimen 

## Packages used in this script 

#library(geomorph) - used for geometric morphometrics data
packageVersion("geomorph")
#4.0.1

#library(morpho) - generate bg-PCA to plot trajectory analyses 
packageVersion("Morpho")
#2.9

#library(phytools) - used for generating PCA phylomorphospace plots and plotting NH test results 
packageVersion("phytools")
#1.0.3

#libraary(scales) - used for plotting transparent points on PCA plots 
packageVersion("scales")
#1.1.1

#library(shape) - nested within the included custom GM_IndvPlot() function 
packageVersion("shape")
#1.4.6

#library(phonTools) - nested within the included custom GM_Indv_Landmark_Outliers () function 
packageVersion("phonTools")  
#0.2.2.1

#library(sp) - nested within the included custom GM_Indv_Landmark_Outliers () function 
packageVersion("sp")  
#1.4.6

#library(plotrix) - nested within the included custom GM_Indv_Landmark_Outliers () function 
packageVersion("plotrix")  
#‘3.8.2’

## Custom functions included at the end of this script 

  # GM_curve_1st_last_remove() - removes the first and last semilandmarks from each individual's curves
  # GM_IndvPlot() - Returns plots of each specimen's landmarks and curves, color coated to facilitate the identification of misplaced or misorderd landmarks or curves 
  # GM_Indv_Landmark_Outliers () - Compares each landmark across specimens and detects and plots outlying landmarks  
  # GM_shape_plot() - plots a representative toe pad shape with light gray silhouette and dark grey lamellae

## My Current R and RStudio Versions 

#check version of R
getRversion()
#R version 4.1.2

#check R studio version
require(rstudioapi)
versionInfo()$version
#R studio version ‘2022.12.0.353’

######
###### Load files ######

#Load geomoprh library
library(geomorph)

#Set working directory
### CHANGE THIS ###
#Change this to the location that you saved the "Final Analyses" folder with all the necessary files in it
#setwd("C://Users/jenny/OneDrive/Desktop/lab images")
setwd("/Users/travishagey/Desktop/Hagey/Science/Projects/Toe Pad Geometric Morphometrics/Parallel gecko toepad evolution/R analyses ")

#load TPS file
dat.raw<-readland.tps("McCann_Shape.tps",specID="imageID", readcurve=T) #read tps
#the specID argument references what specimen names to use
#We didn't use the TPS file's "ID" variable, its values are arbitrary, please ignore
#The readcurve=T argument will treat landmarks in the TPS file under the CURVE headers as semilandmarks
#No scale was measured for these specimens 

#TPS file should load in with no errors
#"150 curve points detected per specimen and are appended to fixed landmarks." 
#This note represents our 10 curves per specimen with an initial 15 points per curve

#load slider file
sliders<-read.table("McCann_Slider.txt")
#This 3 column tab delineated file dictates the relative relationship of each of our semilandmarks
#With each semilandmark in the center column, the left and rigth columns are the two other landmarks or semilandmarks the center semilandmark is connected to on either side


#load category assignment file 
catassign<-read.delim("McCann_Pad_Categories.txt",header=TRUE)


#load ultrametric phylogenetic Squamate tree from Tonini 2016
library(phytools)

# CHANGE this to where ever you have the Tonini 2016 Squamate tree saved
Squamate_tree<-read.tree("/Users/travishagey/Desktop/Hagey/Science/Projects/Phylogenies/Tonini 2016 Squamate tree 5415 ultrametric.tre")


is.ultrametric(Squamate_tree,option=2)

#Check for any polytomies
#rudimentary check for hard polytomies?
is.binary(Squamate_tree)

#hard polytomies (>2 descendants from a node)
length(names(table(Squamate_tree$edge[,1]))[table(Squamate_tree$edge[,1])>2])
#should be 0

#soft polytomy (branch lengths = 0)
any(Squamate_tree$edge.length == 0)
#should be false 

Squamate_tree<-ladderize(Squamate_tree)



###### Remove First and Last Semilandmarks ######

#In order for the slider file to properly correspond with our semilandmarks, we must first remove the first and last semilandmark from each of our curves for all our individuals
#Each curve's first and last landmarks are redundant and overlap with the curve's anchoring landmarks

### MUST LOAD FUNCTION ###
#Run the GM_curve_1st_last_remove() function at the end of this script

to_remove<-GM_curve_1st_last_remove("McCann_Shape.tps")

#We now have a list of the semilandmarks that we need to remove (to_remove) that is compatible with the way geomorph organizes its data
dat.raw<-dat.raw[-to_remove,,] #drops first and last overlapping semilandmarks

#Check to make sure we now have the correct number of landmarks and semilandmarks and specimens
length(dat.raw[,1,1]) #number of total landmarks now
#Should be 141. 11 landmarks, and 13 semilandmarks per curve with 10 curves

length(dat.raw[1,1,]) #number of specimens
#should be 250 individuals


#remove unnecessary variables 
rm(to_remove)

###### Geomorph Procrustes Alignment #####

#Procrustes alignment 
dat.super<-gpagen(dat.raw,curves=sliders, ProcD = F) #The ProcD=F argument denotes the use of bending energy to optimize the location of our semilandmarks


###### Data Quality and Error Identification ####

plotOutliers(dat.super$coord) #Produce an outlier plot to check if any individuals need further attention

#Many of the padless specimens are marked as "outliers" in the above function, but that's likely 
#because they're actually just different than the rest of the specimens. After reviewing them, I do not consider them incorrectly 
#landmarked and so will keep them in the dataset 

#plotAllSpecimens(dat.super$coords) #Overlay all specimens using geomorph function

#You can produce a similar plot, overlaying all specimens, using native function that provide more plotting options 
plot(dat.super$coord[,1,], dat.super$coord[,2,], asp=1, col="grey", pch=19, cex=0.5, xlab=NA, ylab=NA, xaxt="n", yaxt="n")
points(dat.super$consensus, col="red", pch=19, cex=0.5)


#check that the specimens in our trait file match the names in our aligned data file, in the same order
sum(!catassign$species==dimnames(dat.super$coords)[[3]])
#should equal 0


#We use a custom function to plot the unaligned and aligned landmarks and semilandmarks of each individual
### MUST LOAD FUNCTION ###
#Run the GM_IndvPlot() function at the end of this script
#This function produces a multiplot .pdf file for each individual. You must have a destination folder already in place to receive these .pdf plots
#I use the folder "Landmark Diagnostics" within our "Final Analyses" folder to receive these plots  

#Run the function to visualize each individuals landmarks and curves
GM_IndvPlot(gpagen_output = dat.super, TPS_raw = dat.raw, slider_dat = sliders, plot_destination = "Landmark Diagnostics/")

#This function allows users to quickly diagnose individuals in which the landmarks or curves may be out of order
#The function uniquely colors each curve highlighting the order in which the curves should appear. Curves also include an arrow highlighting the order of points
#Since each curve is anchored between two landmarks, if any curves or anchoring landmarks are out of order, this will be evident by misplaced lines connecting them

#Check each landmark individually for outliers across individuals 
GM_Indv_Landmark_Outliers(gpagen_output = dat.super, plot_destination = "Indv Landmark Outliers/")

#This function looks at each landmark individually, checking to see which individuals have landmarks that are significantly different than that same landmark in other specimens 
#Uses both interquartile distance and standard deviation ellipses to determine "significance" 
#We are using this function to find and remedy misplaced landmarks, and so the results need to be considered with 
#a grain of salt. If a specimen's landmark is found to be significantly different, is that because the landmark is misplaced, or because that is just an unusually shaped specimen? 

###### Phylogenetic Tree Prep ####

#Match our data to the species-level phylogeny and modify to phylogeny to work with our data
#Make list of species that we have data for to match them to the phylogeny
Species<-paste(sapply(strsplit(dimnames(dat.super$coords)[[3]], "_"), "[[", 1), sapply(strsplit(dimnames(dat.super$coords)[[3]], "_"), "[[", 2), sep="_")
Species<-unique(Species)

#We have 9 specimens of Aristelliger_cochranae_expectatus, which has been elevated to Aristelliger_expectatus, 
#we don't have any actual Aristelliger_cochranae, and we have one other specimen already named Aristelliger_expectatus
#so we need to update these species names, by just deleting Aristelliger_cochranae from our species list

Species<-Species[!Species=="Aristelliger_cochranae"]
length(unique(Species))
# 94 unique species 

#check to see if all the species we have data for were updated in the phylogeny
#Species[which(!Species %in% Squamate_tree$tip.label)]

#Need to substitute a few names in the phylogeny with closely related species
#In these situations, we assumed the deepest possible relationship between our focal species
Squamate_tree$tip.label[Squamate_tree$tip.label=="Aristelliger_praesignis"]<-"Aristelliger_expectatus"
Squamate_tree$tip.label[Squamate_tree$tip.label=="Sphaerodactylus_dunni"]<-"Sphaerodactylus_inaguae"
Squamate_tree$tip.label[Squamate_tree$tip.label=="Phyllodactylus_galapagensis"]<-"Phyllodactylus_gerrhopygus"
Squamate_tree$tip.label[Squamate_tree$tip.label=="Goggia_braacki"]<-"Goggia_microlepidota"
Squamate_tree$tip.label[Squamate_tree$tip.label=="Goniurosaurus_kuroiwae"]<-"Goniurosaurus_orientalis"
Squamate_tree$tip.label[Squamate_tree$tip.label=="Rhynchoedura_ornata"]<-"Rhynchoedura_eyrensis"

#Misspellings in the phylogeny
Squamate_tree$tip.label[Squamate_tree$tip.label=="Woodworthia_maculatus"]<-"Woodworthia_maculata"

#check to see if all the species we have data for were updated in the phylogeny
Species[which(!Species %in% Squamate_tree$tip.label)]
#should be none, meaning that all of the species for which we have data are in the tree

#Drop the tips of the tree that we don't have data for and rename the tree
Gecko_tree<-drop.tip(Squamate_tree, Squamate_tree$tip.label[which(!Squamate_tree$tip.label %in% Species)])
Gecko_tree #should have 94 tips

# How many genera? 
length(unique(sapply(strsplit(Species, "_"), "[[", 1))) + 1
#Remember that African Cnemaspis have been renamed Ancylodactylus, so we have African Cnemaspis (Ancylodactylus) and Indian Cnemaspis
#that's wny +1
# 56 unique genera

# Species per family 
sp_list<-paste(sapply(strsplit(catassign$species, "_"), "[[", 1), sapply(strsplit(catassign$species, "_"), "[[", 2), sep="_")
sp_list<-cbind(sp_list, catassign$Family)
sp_list[sp_list[,1]=="Aristelliger_cochranae",1]<-"Aristelliger_expectatus"

length(unique(sp_list[sp_list[,2]=="Diplo",1]))
# 17 Diplodactylidae species 
length(unique(sp_list[sp_list[,2]=="Carpho",1]))
# 2 Carphodactylidae species 
length(unique(sp_list[sp_list[,2]=="Eublepharidae",1]))
# 3 Eublepharidae species 
length(unique(sp_list[sp_list[,2]=="Sphaero",1]))
# 9 Sphaerodactylidae species 
length(unique(sp_list[sp_list[,2]=="Phyllo",1]))
# 10 Phyllodactylidae species 
length(unique(sp_list[sp_list[,2]=="Gekko",1]))
# 52 Gekkonidae species 
length(unique(sp_list[sp_list[,2]=="Spheno",1]))
# 1 Sphenodon species 

# Specimens per species 
min(table(sp_list[,1])) # min 1 specimen per species 
max(table(sp_list[,1])) # max 13 specimens per species 
mean(table(sp_list[,1])) # mean 2.65 specimens per species 
sd(table(sp_list[,1])) # sd 2.83 specimens per species 


#we no longer need to total 5000+ tip squamate tree
rm(Squamate_tree)

rm(sp_list)

###### PCA ####
#We could do a true PCA, treating each specimen as an independent measurement, but we are actually more interested 
#in variation across species, and so a between-group PCA, using species as our groups makes more sense 
#visualizing the variation that we're interested in, the variation across species, instead of the variation 
#across specimens. Code for a true PCA is still included below.

#across-species PCA
library(Morpho)
library(scales)

#make list of what species each specimen is
species_list<-paste(sapply(strsplit(dimnames(dat.super$coords)[[3]], "_"), "[[", 1), sapply(strsplit(dimnames(dat.super$coords)[[3]], "_"), "[[", 2), sep="_")

#again, we need to update Aristelliger_cochranae specimens to be Aristelliger_expectatus
species_list[species_list=="Aristelliger_cochranae"]<-"Aristelliger_expectatus"

#do a between group PCA
gPCA_out<-groupPCA(dat.super$coords, groups=species_list, cv=F)
#This is slow and takes a minute. 
#cv=F leaves out the cross-validation step. which is fine since we aren't very interested in how unique each species is
#if cv=F, the function will also throw a warning 
#  "In groupPCA(dat.super$coords, groups = species_list) : group with one entry found - cross-validation will be disabled."
#That just means there's a group with only one specimen in it and so the cross-validation step is skipped anyways. 


cols<-catassign$category
cols[cols=="distal_pad"]<-"orange"
cols[cols=="padless"]<-"red"
cols[cols=="basal_pad"]<-"blue"
cols[cols=="incipient_pad"]<-"green"
names(cols)<-catassign$species


#need species average PC values to add phylomorphospace
#use PCA values to calculate species mean PC values for each axis
Species_PCA<-mat.or.vec(nr = length(Species), nc = 4)
Species_PCA[,]<-NA
rownames(Species_PCA)<-Species
colnames(Species_PCA)<-c("PC1", "PC2", "PC3", "PC4")

for (s in 1:length(Species)){
  #s<-1
  Species_PCA[s,1] <-mean(gPCA_out$Scores[,1][which(species_list==Species[s])])
  Species_PCA[s,2] <-mean(gPCA_out$Scores[,2][which(species_list==Species[s])])
  Species_PCA[s,3] <-mean(gPCA_out$Scores[,3][which(species_list==Species[s])])
  Species_PCA[s,4] <-mean(gPCA_out$Scores[,4][which(species_list==Species[s])])
}

rm(s)

#PC12
pdf(file=file.path("PCA/PC12.pdf"))

plot(gPCA_out$Scores[,1], gPCA_out$Scores[,2], pch=19, col=cols, ylim=rev(range(gPCA_out$Scores[,2])), xlab=NA, ylab=NA, xaxt="n", yaxt="n") #ylim reverses y axis to make it similar to the true PCA plot

phylomorphospace(tree=Gecko_tree, X=cbind(Species_PCA[Gecko_tree$tip.label,"PC1"], Species_PCA[Gecko_tree$tip.label,"PC2"]), add=T, pch=19, axes=F, control=list(col.node=alpha("white", 0)), xlab=NA, ylab=NA, label="off")
points(gPCA_out$Scores[,1], gPCA_out$Scores[,2], pch=19, col=cols)
#text(gPCA_out$Scores[,1], gPCA_out$Scores[,2], rownames(gPCA_out$Scores), cex=0.1)


mtext("Between-Species Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Between-Species Principal Component 2", side=2, line=0.2, cex=1.5)

legend("topleft" , legend=c("Padless", "Incipient pad", "Basal pad", "Distal pad") , pch=19, col=c("red", "green" , "blue" , "orange"), bty="n")

dev.off()


#PC34
pdf(file=file.path("PCA/PC34.pdf"))

plot(gPCA_out$Scores[,3], gPCA_out$Scores[,4], pch=19, col=cols, xlab=NA, ylab=NA, xaxt="n", yaxt="n")
#text(gPCA_out$Scores[,3], gPCA_out$Scores[,4], rownames(gPCA_out$Scores), cex=0.2)

phylomorphospace(tree=Gecko_tree, X=cbind(Species_PCA[Gecko_tree$tip.label,"PC3"], Species_PCA[Gecko_tree$tip.label,"PC4"]), add=T, pch=19, axes=F, control=list(col.node=alpha("white", 0)), xlab=NA, ylab=NA, label="off")
points(gPCA_out$Scores[,3], gPCA_out$Scores[,4], pch=19, col=cols)

mtext("Between-Species Principal Component 3", side=1, line=0.3, cex=1.5)
mtext("Between-Species Principal Component 4", side=2, line=0.2, cex=1.5)

dev.off()

rm(cols)


#determine the amount of variation each PC axis captures
out<-gPCA_out$Variance$cumVar[1:4]

#create proportion of variation plot 
pdf(file=file.path("PCA/PCA variation bar plot.pdf"), width=8, height=7)
par(mar=c(3,5,4,3))
bar<-barplot(out, cex.axis=2.5, ylim=c(0,1), ylab=NA, names.arg=c("PC1", "PC2", "PC3", "PC4"), cex.names =2.5, col="gray")
abline(h=c(0.2, 0.4, 0.6, 0.8), lty=2)
axis(side=1, labels=NA, at=c(0, 10))
barplot(out, cex.axis=2.5, ylim=c(0,1), ylab=NA, names.arg=c("PC1", "PC2", "PC3", "PC4"), cex.names =2.5, col="gray", add=T)
mtext("Cumulative Proportion of Variation", side=3, cex=2.5, line=1)

text(x=bar, y=out, labels = format(round(out, digits = 2), nsmall=2), cex=2, pos=1)

dev.off()

rm(out)
rm(bar)



## visualize projected shape for the min and max of the bgPC axes. 
#bgPC1
proj.shape.sd<- restoreShapes(1*sd(gPCA_out$Scores[,1]), gPCA_out$groupPCs[,1], gPCA_out$Grandmean)
proj.shape.sd.neg<- restoreShapes(-1*sd(gPCA_out$Scores[,1]), gPCA_out$groupPCs[,1], gPCA_out$Grandmean)

proj.shape.sd<-rotate.coords(proj.shape.sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.sd.neg<-rotate.coords(proj.shape.sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("PCA/bgPC1 shape sd.pdf"))
plot(proj.shape.sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd[,1],proj.shape.sd[,2])
mtext(side=1, expression("" %->% "PC1 Maximum"), cex=2)
dev.off()

pdf(file=file.path("PCA/bgPC1 shape -sd.pdf"))
plot(proj.shape.sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd.neg[,1],proj.shape.sd.neg[,2])
mtext(side=1, expression("PC1 Minimum" %<-% ""), cex=2)
dev.off()



#bgPC2
proj.shape.sd<- restoreShapes(1*sd(gPCA_out$Scores[,2]), gPCA_out$groupPCs[,2], gPCA_out$Grandmean)
proj.shape.sd.neg<- restoreShapes(-1*sd(gPCA_out$Scores[,2]), gPCA_out$groupPCs[,2], gPCA_out$Grandmean)

proj.shape.sd<-rotate.coords(proj.shape.sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.sd.neg<-rotate.coords(proj.shape.sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("PCA/bgPC2 shape sd.pdf"))
plot(proj.shape.sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd[,1],proj.shape.sd[,2])
mtext(side=1,  expression("" %down% "PC2 Minimum"), cex=2) #remember the yaxis is inverted on this plot
dev.off()

pdf(file=file.path("PCA/bgPC2 shape -sd.pdf"))
plot(proj.shape.sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd.neg[,1],proj.shape.sd.neg[,2])
mtext(side=1,  expression("" %up% "PC2 Maximum"), cex=2) #remember the yaxis is inverted on this plot
dev.off()



#bgPC3
proj.shape.sd<- restoreShapes(1*sd(gPCA_out$Scores[,3]), gPCA_out$groupPCs[,3], gPCA_out$Grandmean)
proj.shape.sd.neg<- restoreShapes(-1*sd(gPCA_out$Scores[,3]), gPCA_out$groupPCs[,3], gPCA_out$Grandmean)

proj.shape.sd<-rotate.coords(proj.shape.sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.sd.neg<-rotate.coords(proj.shape.sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("PCA/bgPC3 shape sd.pdf"))
plot(proj.shape.sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd[,1],proj.shape.sd[,2])
mtext(side=1, expression("" %->% "PC3 Maximum"), cex=2)
dev.off()

pdf(file=file.path("PCA/bgPC3 shape -sd.pdf"))
plot(proj.shape.sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd.neg[,1],proj.shape.sd.neg[,2])
mtext(side=1, expression("PC3 Minimum" %<-% ""), cex=2)
dev.off()



#bgPC4
proj.shape.sd<- restoreShapes(1*sd(gPCA_out$Scores[,4]), gPCA_out$groupPCs[,4], gPCA_out$Grandmean)
proj.shape.sd.neg<- restoreShapes(-1*sd(gPCA_out$Scores[,4]), gPCA_out$groupPCs[,4], gPCA_out$Grandmean)

proj.shape.sd<-rotate.coords(proj.shape.sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.sd.neg<-rotate.coords(proj.shape.sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("PCA/bgPC4 shape sd.pdf"))
plot(proj.shape.sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd[,1],proj.shape.sd[,2])
mtext(side=1,  expression("" %up% "PC4 Maximum"), cex=2) 

dev.off()

pdf(file=file.path("PCA/bgPC4 shape -sd.pdf"))
plot(proj.shape.sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd.neg[,1],proj.shape.sd.neg[,2])
mtext(side=1,  expression("" %down% "PC4 Minimum"), cex=2) 
dev.off()


rm(gPCA_out)
rm(proj.shape.sd)
rm(proj.shape.sd.neg)
rm(Species_PCA)





#True PC analyses
#This code is here if we wanted to do a specimen-level PCA
if(FALSE){
PCAdat<-gm.prcomp(dat.super$coords) #plotTangentSpace depreciated in geomoprh version 3.3.1

PCscores<-PCAdat$x #PC scores represent each specimen's location on each PC axes


#create PCA plot with species means for each PC axis, the phylogeny connecting the species means, and each species as faint points behind each mean
plotcolors<-catassign$category
plotcolors[plotcolors=="padless"]<-"red"
plotcolors[plotcolors=="basal_pad"]<-"blue"
plotcolors[plotcolors=="incipient_pad"]<-"green"
plotcolors[plotcolors=="distal_pad"]<-"orange"

#use PCA values to calculate species mean values for each axis
Species_PCA<-mat.or.vec(nr = length(Species), nc = 4)
Species_PCA[,]<-NA
rownames(Species_PCA)<-Species
colnames(Species_PCA)<-c("PC1", "PC2", "PC3", "PC4")

for (s in 1:length(Species)){
  #s<-1
  Species_PCA[s,1] <-mean(PCscores[,"Comp1"][which(species_list==Species[s])])
  Species_PCA[s,2] <-mean(PCscores[,"Comp2"][which(species_list==Species[s])])
  Species_PCA[s,3] <-mean(PCscores[,"Comp3"][which(species_list==Species[s])])
  Species_PCA[s,4] <-mean(PCscores[,"Comp4"][which(species_list==Species[s])])
}

#PCA plots 


pdf(file=file.path("PCA/true PCA/PC12.pdf"))

plot(PCscores[,"Comp1"], PCscores[,"Comp2"], pch=19, col=alpha(plotcolors, 0.5), xlab=NA, ylab=NA, xaxt="n", yaxt="n", asp=1)
#text(PCscores[,"Comp1"], PCscores[,"Comp2"], rownames(PCscores), cex=0.2)

points(Species_PCA[,1], Species_PCA[,2], pch=19, col=alpha("white", 0)) #species means are 100% transparent
phylomorphospace(tree=Gecko_tree, X=cbind(Species_PCA[Gecko_tree$tip.label,"PC1"], Species_PCA[Gecko_tree$tip.label,"PC2"]), add=T, pch=19, axes=F, control=list(col.node=alpha("white", 0)), xlab=NA, ylab=NA, label="off")
points(PCscores[,"Comp1"], PCscores[,"Comp2"], pch=19, col=alpha(plotcolors, 0.5))

mtext("Principal Component 1", side=1, line=0.2, cex=1.5)
mtext("Principal Component 2", side=2, line=0.2, cex=1.5)

legend("bottomleft" , legend=c("Padless", "Incipient pad", "Basal pad", "Distal pad") , pch=19, col=c("red", "green" , "blue" , "orange"))

dev.off()



pdf(file=file.path("PCA/true PCA/PC34.pdf"))

plot(PCscores[,"Comp3"], PCscores[,"Comp4"], pch=19, col=alpha(plotcolors, 0.5), xlab=NA, ylab=NA, xaxt="n", yaxt="n", asp=1)

points(Species_PCA[,3], Species_PCA[,4], pch=19, col=alpha("white", 0)) #species means are 100% transparent
phylomorphospace(tree=Gecko_tree, X=cbind(Species_PCA[Gecko_tree$tip.label,"PC3"], Species_PCA[Gecko_tree$tip.label,"PC4"]), add=T, pch=19, axes=F, control=list(col.node=alpha("white", 0)), xlab=NA, ylab=NA, label="off")
points(PCscores[,"Comp3"], PCscores[,"Comp4"], pch=19, col=alpha(plotcolors, 0.5))


mtext("Principal Component 3", side=1, line=0.2, cex=1.5)
mtext("Principal Component 4", side=2, line=0.2, cex=1.5)

#text(PCscores[,"Comp3"], PCscores[,"Comp4"], rownames(PCscores), cex=0.5)
#legend("bottomleft" , legend=c("Padless", "Incipient pad", "Basal pad", "Distal pad") , pch=19, col=c("red", "green" , "blue" , "orange"))

dev.off()


#determine the amount of variation each PC axis captures
out<-summary(PCAdat)
out$PC.summary[3,1:4]


#create proportion of variation plot 
pdf(file=file.path("PCA/true PCA/PCA variation bar plot.pdf"))
par(mar=c(3,5,4,5))
barplot(as.matrix(out$PC.summary[3, 1:4]), cex.axis=2.5, ylim=c(0,1), ylab=NA, names.arg=c("PC1", "PC2", "PC3", "PC4"), cex.names =2.5, col="gray")
abline(h=c(0.2, 0.4, 0.6, 0.8), lty=2)
axis(side=1, labels=NA, at=c(0, 10))
barplot(as.matrix(out$PC.summary[3, 1:4]), cex.axis=2.5, ylim=c(0,1), ylab=NA, names.arg=c("PC1", "PC2", "PC3", "PC4"), cex.names =2.5, col="gray", add = T)
mtext("Cumulative Proportion\nof Variation", side=3, cex=2.5, line=-2)

dev.off()


#plot estimated shape for each PC axis extreme
#rotate image 90 degrees
PC1min<-rotate.coords(PCAdat$shapes$shapes.comp1$min, type="rotateCC")
PC1max<-rotate.coords(PCAdat$shapes$shapes.comp1$max, type="rotateCC")
PC2min<-rotate.coords(PCAdat$shapes$shapes.comp2$min, type="rotateCC")
PC2max<-rotate.coords(PCAdat$shapes$shapes.comp2$max, type="rotateCC")
PC3min<-rotate.coords(PCAdat$shapes$shapes.comp3$min, type="rotateCC")
PC3max<-rotate.coords(PCAdat$shapes$shapes.comp3$max, type="rotateCC")
PC4min<-rotate.coords(PCAdat$shapes$shapes.comp4$min, type="rotateCC")
PC4max<-rotate.coords(PCAdat$shapes$shapes.comp4$max, type="rotateCC")


#design estimated shape plot and plot on pdf

pdf(file=file.path("PCA/true PCA/PC1minshape.pdf"))
plot(PC1min, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(PC1min[,1],PC1min[,2])
mtext(side=1, expression("PC1 Minimum" %<-% ""), cex=2)
dev.off()

pdf(file=file.path("PCA/true PCA/PC1maxshape.pdf"))
plot(PC1max, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(PC1max[,1],PC1max[,2])
mtext(side=1, expression("" %->% "PC1 Maximum"), cex=2)
dev.off()

pdf(file=file.path("PCA/true PCA/PC2minshape.pdf"))
plot(PC2min, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(PC2min[,1],PC2min[,2])
mtext(side=1,  expression("" %down% "PC2 Minimum"), cex=2)
dev.off()

pdf(file=file.path("PCA/true PCA/PC2maxshape.pdf"))
plot(PC2max, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(PC2max[,1],PC2max[,2])
mtext(side=1,  expression("" %up% "PC2 Maximum"), cex=2)
dev.off()

pdf(file=file.path("PCA/true PCA/PC3minshape.pdf"))
plot(PC3min, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(PC3min[,1],PC3min[,2])
mtext(side=1, expression("PC3 Minimum" %<-% ""), cex=2)
dev.off()

pdf(file=file.path("PCA/true PCA/PC3maxshape.pdf"))
plot(PC3max, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(PC3max[,1],PC3max[,2])
mtext(side=1, expression("" %->% "PC3 Maximum"), cex=2)
dev.off()

pdf(file=file.path("PCA/true PCA/PC4minshape.pdf"))
plot(PC4min, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(PC4min[,1],PC4min[,2])
mtext(side=1,  expression("" %down% "PC4 Minimum"), cex=2)
dev.off()

pdf(file=file.path("PCA/true PCA/PC4maxshape.pdf"))
plot(PC4max, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(PC4max[,1],PC4max[,2])
mtext(side=1,  expression("" %up% "PC4 Maximum"), cex=2)
dev.off()


#rm unnecessary variables 
rm(out)
rm(PC1max)
rm(PC1min)
rm(PC2max)
rm(PC2min)
rm(PC3max)
rm(PC3min)
rm(PC4max)
rm(PC4min)
rm(PCAdat)
rm(PCscores)
rm(Species_PCA)
rm(plotcolors)
rm(s)

}


###### Clustering ####
#Before we can investigate patterns of parallel toe pad evolution, we first need to ask if our a priori assigned toe pad classes represent 
#real patterns in trait space.
#Our previous PCA suggests our assigned toe pad classes are real, but we will conduct additional analyses to confirm this
#To first investigate how "real" our a priori assigned toe pad classes are, we can correlate assigned pad class with shape
#to ask if shape significantly correlates with our assigned classes. 

#need to calculate species averages for each species
Species_Mshape<-array(dim = c(141, 2, length(Species)))
dimnames(Species_Mshape)[[3]]<-Species

for (s in 1:length(Species)){
  #s<-1
  Species_Mshape[,,s]<-mshape(dat.super$coords[,,grep(x=dimnames(dat.super$coords)[[3]], pattern=Species[s])])
  }

#since all 4 of the "Aristelliger_cochranae_expectatus" are really "Aristelliger_expectatus", but have different names, the above averaging didn't include them
#have to average Aristelliger_expectatus by hand

Species_Mshape[,,"Aristelliger_expectatus"]<-mshape(dat.super$coords[,,grep(dimnames(dat.super$coords)[[3]], pattern="Aristelliger")])

rm(s)

species_cat<-rep(NA, length(Species))
names(species_cat)<-Species

for (s in 1:length(Species)){
  #s<-23
  species_cat[s]<-catassign$category[grep(x=catassign$species, pattern=Species[s])][1]
}


gdf <- geomorph.data.frame(Coords=Species_Mshape, Cats = species_cat, phylo=Gecko_tree)
fit<-procD.pgls(Coords~Cats, phy=phylo, data = gdf) 
summary(fit) 
fit$aov.table["Cats", "Pr(>F)"] #pval = 0.001 
#so yes, toe pad category can significantly predict shape

rm(gdf)
rm(fit)

#also conduct a bgPCA to see how well specimens can be classified 

gPCA_out<-groupPCA(dat.super$coords, groups=catassign$category)
gPCA_out
gPCA_out$probs

#our bgPCA classified specimens into the correct class 92% of the time
#distal vs basal - these were significantly different (p=0.01)
#incipient vs basal, incipient vs distal, padless vs basal, and padless vs distal were all sig (p=0.001)
#padless vs incipient was not significantly different (p=0.13)



#Clustering analyses 
#Can also do clustering analyses, grouping specimens based on their shape similarities 
#this results in a dendrogram, superficially similar to a phylogeny, but instead of showing evolutionary relationships,
#a dendrogram highlights similarities in whatever variable you provide, in our case, it is shape

#We will primarily use on Ward's (1963) clustering criterion to build our dendrogram, which aims to find compact, spherical clusters
#but there are other options like the complete linkage method which finds similar clusters based on clade's max differences.
#The "average" or UPGMA method defines the differences between clades as the difference in their average value. 


dat.clustering<-dat.raw

dat.clustering<-gpagen(dat.clustering,curves=sliders, ProcD = F, verbose = T) #The ProcD=F argument denotes the use of bending energy to optimize the location of our semilandmarks
#need to do another alignment with verbose=T to get a distance matrix 

cluster.dendro.ward<-hclust(dat.clustering$procD, method = "ward.D2")
cluster.dendro.ward<-as.phylo(cluster.dendro.ward)

cols<-catassign$category
cols[cols=="distal_pad"]<-"orange"
cols[cols=="padless"]<-"red"
cols[cols=="basal_pad"]<-"blue"
cols[cols=="incipient_pad"]<-"green"
names(cols)<-catassign$species

fams<-catassign$Family
fams[fams=="Gekko"]<- "darkorchid1"
fams[fams=="Eublepharidae"]<-"hotpink2"
fams[fams=="Sphaero"]<-"aquamarine4"
fams[fams=="Carpho"]<-"gold1"
fams[fams=="Phyllo"]<-"tan"
fams[fams=="Diplo"]<-"ivory4"
fams[fams=="Spheno"]<-"salmon4"
names(fams)<-catassign$species


#make a set of branch colors based on pad category
branches<-(rep("black", nrow(cluster.dendro.ward$edge)))

for(t in 1:length(cluster.dendro.ward$tip.label)){
  #t<-1
  c<-cols[catassign$species==cluster.dendro.ward$tip.label[t]]
  branches[which(cluster.dendro.ward$edge[,2]==t)]<-c
}

pdf(file=file.path("cluster/dendroWard.pdf"), width=5, height = 6)
par(mar=c(0,0,1.5,0))

plot(cluster.dendro.ward, cex=0.2, tip.color=fams, edge.color=branches)

legend("topleft", inset=c(0.,-0.05), cex=0.8, legend = c(expression(bold("Branch colors: Toe Class")), "Padless", "Incipient", "Distal", "Basal"), bty = "n", text.col = c("black", "red", "green", "orange", "blue"), xpd = T)
legend("topleft", inset=c(0.,0.1), cex=0.8, legend = c(expression(bold("Tip colors: Families")), "Sphenodon", "Carphodactylidae", "Diplodactylidae", "Eublepharidae", "Sphaerodactylidae", "Phyllodactylidae", "Gekkonidae"), bty = "n", text.col = c("black", "salmon4", "gold1", "ivory4", "hotpink2", "aquamarine4", "tan", "darkorchid1" ), xpd=T)

dev.off()


#Can also check out other clustering methods including the complete method or the average method 
if(FALSE){

  #Complete method
  cluster.dendro.comp<-hclust(dat.clustering$procD, method = "complete")
  cluster.dendro.comp<-as.phylo(cluster.dendro.comp)

  #make a set of branch colors based on pad category
  branches<-(rep("black", nrow(cluster.dendro.comp$edge)))

  for(t in 1:length(cluster.dendro.comp$tip.label)){
    #t<-1
    c<-cols[catassign$species==cluster.dendro.comp$tip.label[t]]
    branches[which(cluster.dendro.comp$edge[,2]==t)]<-c
  }

  pdf(file=file.path("cluster/other methods/dendroComp.pdf"), width=5, height = 6)
  par(mar=c(0,0,1.5,0))

  plot(cluster.dendro.comp, cex=0.2, tip.color=fams, main="Complete", edge.color=branches)

  legend("topleft", inset=c(0.,-0.05), cex=0.8, legend = c(expression(bold("Branch colors: Toe Class")), "Padless", "Incipient", "Distal", "Basal"), bty = "n", text.col = c("black", "red", "green", "orange", "blue"), xpd = T)
  legend("topleft", inset=c(0.,0.1), cex=0.8, legend = c(expression(bold("Tip colors: Families")), "Sphenodon", "Carphodactylidae", "Diplodactylidae", "Eublepharidae", "Sphaerodactylidae", "Phyllodactylidae", "Gekkonidae"), bty = "n", text.col = c("black", "salmon4", "gold1", "ivory4", "hotpink2", "aquamarine4", "tan", "darkorchid1" ), xpd=T)

  dev.off()



  #average method
  cluster.dendro.ave<-hclust(dat.clustering$procD, method = "ave")
  cluster.dendro.ave<-as.phylo(cluster.dendro.ave)

  #make a set of branch colors based on pad category
  branches<-(rep("black", nrow(cluster.dendro.ave$edge)))

  for(t in 1:length(cluster.dendro.ave$tip.label)){
    #t<-1
    c<-cols[catassign$species==cluster.dendro.ave$tip.label[t]]
    branches[which(cluster.dendro.ave$edge[,2]==t)]<-c
  }


  pdf(file=file.path("cluster/other methods/dendroAve.pdf"), width=5, height = 6)
  par(mar=c(0,0,1.5,0))

  plot(cluster.dendro.ave, cex=0.2, tip.color=fams, main="UPGMA", edge.color=branches)

  legend("topleft", inset=c(0.,-0.05), cex=0.8, legend = c(expression(bold("Branch colors: Toe Class")), "Padless", "Incipient", "Distal", "Basal"), bty = "n", text.col = c("black", "red", "green", "orange", "blue"), xpd = T)
  legend("topleft", inset=c(0.,0.1), cex=0.8, legend = c(expression(bold("Tip colors: Families")), "Sphenodon", "Carphodactylidae", "Diplodactylidae", "Eublepharidae", "Sphaerodactylidae", "Phyllodactylidae", "Gekkonidae"), bty = "n", text.col = c("black", "salmon4", "gold1", "ivory4", "hotpink2", "aquamarine4", "tan", "darkorchid1" ), xpd=T)

  dev.off()

  rm(cluster.dendro.ave)
  rm(cluster.dendro.comp)
}


rm(dat.clustering)
rm(gPCA_out)
rm(cluster.dendro.ward)
rm(branches)
rm(c)
rm(cols)
rm(fams)
rm(s)
rm(t)



###### Padless ####

#We'll now consider the variation within the padless class, which spans all 6 focal gecko families plus Sphenodon.
#Since this is the morphological class that all other toe pad forms presumably evolved from, we wanted to quantify 
#any pre-existing variation, and ask if it has phylogenetic signal. Lineage-specific differences in padless taxa may have 
#preemptively affected different lineages' subsequent evolution of toe pads

#We are interested in how morphologically universal or consistent "padless-ness" is
#We included shape data for padless families of geckos (Carphodactylidae and Eublepharidae), 
#padless examples within Diplodactylidae, Gekkonidae, Phyllodactylidae, and Sphaerodactylidae,
#and Sphenodon, the sister taxa to Squamata


#check that catassign species names match dat.super$coords names
catassign$species==dimnames(dat.super$coords)[[3]] #should be good

#Is there interesting patterns of variation across padless taxa? 
dat.padless.shape<-dat.super$coords[,,catassign$category=="padless"]

#Correlate padless shape with family (PGLS)
padless_list<-paste(sapply(strsplit(dimnames(dat.padless.shape)[[3]], "_"), "[[", 1), sapply(strsplit(dimnames(dat.padless.shape)[[3]], "_"), "[[", 2), sep="_")

#Use species average shape for pgls
padless_species_means<-Species_Mshape[,,unique(padless_list)]

#make a padless phylo
padless_tree<-drop.tip(Gecko_tree, Gecko_tree$tip.label[!Gecko_tree$tip.label %in% unique(padless_list)])


#need to know each species' family to use as independent variable in PGLS 
padless_fams<-rep(NA, length(unique(padless_list)))
names(padless_fams)<-unique(padless_list)

for(f in 1:length(padless_fams)){
  #f<-2
  padless_fams[f]<-catassign$Family[grep(pattern=unique(padless_list)[f], x=catassign$species)] [1] #should all be the same, just take the 1st one
}

rm(f)

gdf <- geomorph.data.frame(Coords=padless_species_means, Families = padless_fams, phylo=padless_tree)
fit<-procD.pgls(Coords~Families, phy=phylo, data = gdf) 
summary(fit) 
PGLS_out<-fit$aov.table["Families", "Pr(>F)"] #pval = 0.8
PGLS_out

#when only considering padless species, family does not correlate with shape

rm(fit)
rm(gdf)


#Conduct a species level PCA to identify major axes of variation and how/if padless toes vary across families
PCAdat<-groupPCA(dat.padless.shape, groups=padless_list, cv=F)
#This is slow and takes a minute. 
#cv=F leaves out the cross-validation step. which is fine since we aren't very interested in how unique each species is
#if cv=F, the function will also throw a warning 
#  warning message "group with one entry found - cross-validation will be disabled."
#That just means there's a group with only one specimen in it and so the cross-validation step is skipped anyways. 

PCscores<-PCAdat$Scores

#create PCA plot with families colored, with the phylogeny connecting species means, and each specimen as faint point
plotcolors<-catassign$Family[catassign$category=="padless"]

#based on: brewer.pal(7, "YlOrRd") from the library(RColorBrewer) package
plotcolors[plotcolors=="Spheno"]<-"#B10026"     #brownish red
plotcolors[plotcolors=="Carpho"]<-"#E31A1C"     #red
plotcolors[plotcolors=="Diplo"]<-"#FC4E2A"      #red/orange
plotcolors[plotcolors=="Eublepharidae"]<-"#FD8D3C" #orange
plotcolors[plotcolors=="Sphaero"]<-"#FEB24C"    #orange/yellow
plotcolors[plotcolors=="Phyllo"]<-"#FED976"     #soft yellow
plotcolors[plotcolors=="Gekko"]<-"#FFFFB2"      #pastel yellow
names(plotcolors)<-catassign$species[catassign$category=="padless"]

plotcolors<-cbind(catassign$Family[catassign$category=="padless"], plotcolors)
colnames(plotcolors)[1]<-"Fam"

fam.cols<-unique(padless_fams)
fam.cols[fam.cols=="Spheno"]<-"#B10026"     #brownish red
fam.cols[fam.cols=="Carpho"]<-"#E31A1C"     #red
fam.cols[fam.cols=="Diplo"]<-"#FC4E2A"      #red/orange
fam.cols[fam.cols=="Eublepharidae"]<-"#FD8D3C" #orange
fam.cols[fam.cols=="Sphaero"]<-"#FEB24C"    #orange/yellow
fam.cols[fam.cols=="Phyllo"]<-"#FED976"     #soft yellow
fam.cols[fam.cols=="Gekko"]<-"#FFFFB2"    #pastel yellow
names(fam.cols)<-unique(padless_fams)



#PCA plots 
pdf(file=file.path("padless/PC12.pdf"), width=5, height = 5)
par(mar=c(3,3,3,1)) 

plot(PCscores[,1], PCscores[,2], pch=19, col=alpha("gray", 0.5), xlab=NA, ylab=NA, xaxt="n", yaxt="n")
#text(PCscores[,1], PCscores[,2], rownames(PCscores), cex=0.2)

#plot convex hull around each family's individuals 
hpts<-chull(PCscores[c(which(plotcolors[,1]=="Spheno")), 1], y = PCscores[c(which(plotcolors[,1]=="Spheno")), 2]) #get list of outer forest points
hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
lines(PCscores[c(which(plotcolors[,1]=="Spheno")), 1][hpts], PCscores[c(which(plotcolors[,1]=="Spheno")), 2][hpts], col=alpha(fam.cols["Spheno"], 0.5)) #plot the lines

#plot convex hull around each family's individuals 
hpts<-chull(PCscores[c(which(plotcolors[,1]=="Carpho")), 1], y = PCscores[c(which(plotcolors[,1]=="Carpho")), 2]) #get list of outer forest points
hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
lines(PCscores[c(which(plotcolors[,1]=="Carpho")), 1][hpts], PCscores[c(which(plotcolors[,1]=="Carpho")), 2][hpts], col=alpha(fam.cols["Carpho"], 0.5)) #plot the lines

#plot convex hull around each family's individuals 
hpts<-chull(PCscores[c(which(plotcolors[,1]=="Diplo")), 1], y = PCscores[c(which(plotcolors[,1]=="Diplo")), 2]) #get list of outer forest points
hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
lines(PCscores[c(which(plotcolors[,1]=="Diplo")), 1][hpts], PCscores[c(which(plotcolors[,1]=="Diplo")), 2][hpts], col=alpha(fam.cols["Diplo"], 0.5)) #plot the lines

#plot convex hull around each family's individuals 
hpts<-chull(PCscores[c(which(plotcolors[,1]=="Eublepharidae")), 1], y = PCscores[c(which(plotcolors[,1]=="Eublepharidae")), 2]) #get list of outer forest points
hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
lines(PCscores[c(which(plotcolors[,1]=="Eublepharidae")), 1][hpts], PCscores[c(which(plotcolors[,1]=="Eublepharidae")), 2][hpts], col=alpha(fam.cols["Eublepharidae"], 0.5)) #plot the lines

#plot convex hull around each family's individuals 
hpts<-chull(PCscores[c(which(plotcolors[,1]=="Sphaero")), 1], y = PCscores[c(which(plotcolors[,1]=="Sphaero")), 2]) #get list of outer forest points
hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
lines(PCscores[c(which(plotcolors[,1]=="Sphaero")), 1][hpts], PCscores[c(which(plotcolors[,1]=="Sphaero")), 2][hpts], col=alpha(fam.cols["Sphaero"], 0.5)) #plot the lines

#plot convex hull around each family's individuals 
hpts<-chull(PCscores[c(which(plotcolors[,1]=="Phyllo")), 1], y = PCscores[c(which(plotcolors[,1]=="Phyllo")), 2]) #get list of outer forest points
hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
lines(PCscores[c(which(plotcolors[,1]=="Phyllo")), 1][hpts], PCscores[c(which(plotcolors[,1]=="Phyllo")), 2][hpts], col=alpha(fam.cols["Phyllo"], 0.5)) #plot the lines

#plot convex hull around each family's individuals 
hpts<-chull(PCscores[c(which(plotcolors[,1]=="Gekko")), 1], y = PCscores[c(which(plotcolors[,1]=="Gekko")), 2]) #get list of outer forest points
hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
lines(PCscores[c(which(plotcolors[,1]=="Gekko")), 1][hpts], PCscores[c(which(plotcolors[,1]=="Gekko")), 2][hpts], col=alpha("grey", 0.5)) #plot the lines
lines(PCscores[c(which(plotcolors[,1]=="Gekko")), 1][hpts], PCscores[c(which(plotcolors[,1]=="Gekko")), 2][hpts], col=alpha(fam.cols["Gekko"], 0.5)) #plot the lines



#calculate species means in PC plot

s.means<-rep(NA, length(unique(padless_tree$tip.label)))
s.means<-cbind(s.means, s.means)
colnames(s.means)<-c("PC1", "PC2")
rownames(s.means)<-unique(padless_tree$tip.label)

for(s in unique(padless_tree$tip.label)){
  #s<-"Sphenodon_punctatus"
  out<-PCscores[grep(rownames(PCscores), pattern=s),1] #pc1
  s.means[s,"PC1"]<-mean(out)
  
  out<-PCscores[grep(rownames(PCscores), pattern=s),2] #pc2
  s.means[s,"PC2"]<-mean(out)
}

#plot(padless_tree, no.margin=T)

points(PCscores[,1], PCscores[,2], pch=19, col=alpha(plotcolors[,2], 0.3)) #add individual specimens

phylomorphospace(tree=padless_tree, X=s.means, add=T, pch=19, axes=F, control=list(col.node=alpha("white", 0)), xlab=NA, ylab=NA, label="off")

s.cols<-padless_fams[rownames(s.means)]
for(s in 1:length(s.cols)){
  #s<-1
  s.cols[s]<-fam.cols[s.cols[s]]
}

points(s.means, pch=21, bg=alpha(s.cols,1)) #add family averages

mtext("Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Principal Component 2", side=2, line=0.2, cex=1.5)

legend("topleft" , legend=rep(NA, length(fam.cols)), cex=0.7, pch=21, pt.bg=c("gray"), bty = "n")
legend("topleft" , cex=0.7, legend=c("Sphenodon", "Carphodactylidae", "Diplodactylidae", "Eublepharidae", "Sphaerodactylidae", "Phyllodactylidae", "Gekkonidae") , pch=21, pt.bg=alpha(c("#B10026", "#E31A1C" , "#FC4E2A" , "#FD8D3C", "#FEB24C", "#FED976", "#FFFFB2"),0.5), bty = "n")

mtext(side=1, line=-1.1, adj = 0.01, text=paste("PGLS: shape ~ family: p = ", round(PGLS_out,3), sep=""), cex=0.7)

dev.off()




#PC 3 vs 4 (didn't include)
if(FALSE){
  pdf(file=file.path("padless/PC34.pdf"), width=5, height = 5)
  par(mar=c(3,3,3,1)) 
  
  plot(PCscores[,3], PCscores[,4], pch=19, col=alpha("gray", 0.5), xlab=NA, ylab=NA, xaxt="n", yaxt="n")
  #text(PCscores[,3], PCscores[,4], rownames(PCscores), cex=0.2)

  #plot convex hull around each family's individuals 
  hpts<-chull(PCscores[c(which(plotcolors[,1]=="Spheno")), 3], y = PCscores[c(which(plotcolors[,1]=="Spheno")), 4]) #get list of outer forest points
  hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
  lines(PCscores[c(which(plotcolors[,1]=="Spheno")), 3][hpts], PCscores[c(which(plotcolors[,1]=="Spheno")), 4][hpts], col=alpha(fam.cols["Spheno"], 0.5)) #plot the lines

  #plot convex hull around each family's individuals 
  hpts<-chull(PCscores[c(which(plotcolors[,1]=="Carpho")), 3], y = PCscores[c(which(plotcolors[,1]=="Carpho")), 4]) #get list of outer forest points
  hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
  lines(PCscores[c(which(plotcolors[,1]=="Carpho")), 3][hpts], PCscores[c(which(plotcolors[,1]=="Carpho")), 4][hpts], col=alpha(fam.cols["Carpho"], 0.5)) #plot the lines

  #plot convex hull around each family's individuals 
  hpts<-chull(PCscores[c(which(plotcolors[,1]=="Diplo")), 3], y = PCscores[c(which(plotcolors[,1]=="Diplo")), 4]) #get list of outer forest points
  hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
  lines(PCscores[c(which(plotcolors[,1]=="Diplo")), 3][hpts], PCscores[c(which(plotcolors[,1]=="Diplo")), 4][hpts], col=alpha(fam.cols["Diplo"], 0.5)) #plot the lines

  #plot convex hull around each family's individuals 
  hpts<-chull(PCscores[c(which(plotcolors[,1]=="Eublepharidae")), 3], y = PCscores[c(which(plotcolors[,1]=="Eublepharidae")), 4]) #get list of outer forest points
  hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
  lines(PCscores[c(which(plotcolors[,1]=="Eublepharidae")), 3][hpts], PCscores[c(which(plotcolors[,1]=="Eublepharidae")), 4][hpts], col=alpha(fam.cols["Eublepharidae"], 0.5)) #plot the lines

  #plot convex hull around each family's individuals 
  hpts<-chull(PCscores[c(which(plotcolors[,1]=="Sphaero")), 3], y = PCscores[c(which(plotcolors[,1]=="Sphaero")), 4]) #get list of outer forest points
  hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
  lines(PCscores[c(which(plotcolors[,1]=="Sphaero")), 3][hpts], PCscores[c(which(plotcolors[,1]=="Sphaero")), 4][hpts], col=alpha(fam.cols["Sphaero"], 0.5)) #plot the lines

  #plot convex hull around each family's individuals 
  hpts<-chull(PCscores[c(which(plotcolors[,1]=="Phyllo")), 3], y = PCscores[c(which(plotcolors[,1]=="Phyllo")), 4]) #get list of outer forest points
  hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
  lines(PCscores[c(which(plotcolors[,1]=="Phyllo")), 3][hpts], PCscores[c(which(plotcolors[,1]=="Phyllo")), 4][hpts], col=alpha(fam.cols["Phyllo"], 0.5)) #plot the lines

  #plot convex hull around each family's individuals 
  hpts<-chull(PCscores[c(which(plotcolors[,1]=="Gekko")), 3], y = PCscores[c(which(plotcolors[,1]=="Gekko")), 4]) #get list of outer forest points
  hpts<-c(hpts, hpts[1]) #duplicate first point to close the polygon
  lines(PCscores[c(which(plotcolors[,1]=="Gekko")), 3][hpts], PCscores[c(which(plotcolors[,1]=="Gekko")), 4][hpts], col=alpha("grey", 0.5)) #plot the lines
  lines(PCscores[c(which(plotcolors[,1]=="Gekko")), 3][hpts], PCscores[c(which(plotcolors[,1]=="Gekko")), 4][hpts], col=alpha(fam.cols["Gekko"], 0.5)) #plot the lines

  phylomorphospace(tree=fam_tree, X=cbind(Family_PCA[,"PC3"], Family_PCA[,"PC4"]), add=T, pch=19, axes=F, control=list(col.node=alpha("white", 0)), xlab=NA, ylab=NA, label="off")
  points(Family_PCA[,3], Family_PCA[,4], pch=21, bg="gray") #add family averages
  points(Family_PCA[,3], Family_PCA[,4], pch=21, bg=alpha(fam.cols,0.5)) #add family averages
  points(PCscores[,3], PCscores[,4], pch=19, col=alpha(plotcolors[,2], 0.5)) #add individual specimens

  mtext("Principal Component 3", side=1, line=0.3, cex=1.5)
  mtext("Principal Component 4", side=2, line=0.2, cex=1.5)

  dev.off()

}


#determine the amount of variation each PC axis captures
out<-PCAdat$Variance$cumVar[1:2]
out #0.8423808 0.9119964


#create proportion of variation plot 
pdf(file=file.path("padless/PCA variation bar plot.pdf"))
par(mar=c(3,7,4,5))
bar<-barplot(out, cex.axis=2.5, ylim=c(0,1), ylab=NA, names.arg=c("PC1", "PC2"), cex.names =2.5, col="gray")
abline(h=c(0.2, 0.4, 0.6, 0.8), lty=2)
axis(side=1, labels=NA, at=c(0, 10))
barplot(out, cex.axis=2.5, ylim=c(0,1), ylab=NA, names.arg=c("PC1", "PC2"), cex.names =2.5, col="gray", add=T)
mtext("Cumulative Proportion\nof Variation", side=3, cex=2.5, line=-1)

text(x=bar, y=out, labels = format(round(out, digits = 2), nsmall=2), cex=2, pos=1)

dev.off()

rm(bar)

## visualize projected shape for the min and max of the bgPC axes. 
#bgPC1
proj.shape.sd<- restoreShapes(1*sd(PCAdat$Scores[,1]), PCAdat$groupPCs[,1], PCAdat$Grandmean)
proj.shape.sd.neg<- restoreShapes(-1*sd(PCAdat$Scores[,1]), PCAdat$groupPCs[,1], PCAdat$Grandmean)

proj.shape.sd<-rotate.coords(proj.shape.sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.sd.neg<-rotate.coords(proj.shape.sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("padless/bgPC1 shape sd.pdf"))
plot(proj.shape.sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd[,1],proj.shape.sd[,2])
mtext(side=1, expression("" %->% "PC1 Maximum"), cex=2)
dev.off()

pdf(file=file.path("padless/bgPC1 shape -sd.pdf"))
plot(proj.shape.sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd.neg[,1],proj.shape.sd.neg[,2])
mtext(side=1, expression("PC1 Minimum" %<-% ""), cex=2)
dev.off()



#bgPC2
proj.shape.sd<- restoreShapes(1*sd(PCAdat$Scores[,2]), PCAdat$groupPCs[,2], PCAdat$Grandmean)
proj.shape.sd.neg<- restoreShapes(-1*sd(PCAdat$Scores[,2]), PCAdat$groupPCs[,2], PCAdat$Grandmean)

proj.shape.sd<-rotate.coords(proj.shape.sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.sd.neg<-rotate.coords(proj.shape.sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("padless/bgPC2 shape sd.pdf"))
plot(proj.shape.sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd[,1],proj.shape.sd[,2])
mtext(side=1,  expression("" %up% "PC2 Maximum"), cex=2) 
dev.off()

pdf(file=file.path("padless/bgPC2 shape -sd.pdf"))
plot(proj.shape.sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd.neg[,1],proj.shape.sd.neg[,2])
mtext(side=1,  expression("" %down% "PC2 Minimum"), cex=2) 
dev.off()



#bgPC3
proj.shape.sd<- restoreShapes(1*sd(PCAdat$Scores[,3]), PCAdat$groupPCs[,3], PCAdat$Grandmean)
proj.shape.sd.neg<- restoreShapes(-1*sd(PCAdat$Scores[,3]), PCAdat$groupPCs[,3], PCAdat$Grandmean)

proj.shape.sd<-rotate.coords(proj.shape.sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.sd.neg<-rotate.coords(proj.shape.sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("padless/bgPC3 shape sd.pdf"))
plot(proj.shape.sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd[,1],proj.shape.sd[,2])
mtext(side=1, expression("" %->% "PC3 Maximum"), cex=2)
dev.off()

pdf(file=file.path("padless/bgPC3 shape -sd.pdf"))
plot(proj.shape.sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd.neg[,1],proj.shape.sd.neg[,2])
mtext(side=1, expression("PC3 Minimum" %<-% ""), cex=2)
dev.off()



#bgPC4
proj.shape.sd<- restoreShapes(1*sd(PCAdat$Scores[,4]), PCAdat$groupPCs[,4], PCAdat$Grandmean)
proj.shape.sd.neg<- restoreShapes(-1*sd(PCAdat$Scores[,4]), PCAdat$groupPCs[,4], PCAdat$Grandmean)

proj.shape.sd<-rotate.coords(proj.shape.sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.sd.neg<-rotate.coords(proj.shape.sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("padless/bgPC4 shape sd.pdf"))
plot(proj.shape.sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd[,1],proj.shape.sd[,2])
mtext(side=1,  expression("" %up% "PC4 Maximum"), cex=2) 

dev.off()

pdf(file=file.path("padless/bgPC4 shape -sd.pdf"))
plot(proj.shape.sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.sd.neg[,1],proj.shape.sd.neg[,2])
mtext(side=1,  expression("" %down% "PC4 Minimum"), cex=2) 
dev.off()




#rm unnecessary variables 
rm(out)
rm(PCAdat)
rm(PCscores)
rm(hpts)
rm(padless_tree)
rm(plotcolors)
rm(dat.padless.shape)
rm(fam.cols)
rm(padless_fams)
rm(padless_list)
rm(padless_species_means)
rm(PGLS_out)
rm(s)
rm(s.cols)
rm(s.means)
rm(proj.shape.sd)
rm(proj.shape.sd.neg)


###### Trajectory Analyses ####
#We then compared the distance and angle in Procrustes space (ie bgPCA space), between different toe pad categories 
#We are interested in parallel evolution across families of geckos, and so we will compare the differences 
#between toe pad classes and how these differences vary across families. For example, the transition between padlessness 
#and having basal pads. Is this difference similar across families? Its worth noting that we aren't necessarily interested 
#in the direction of the transition (i.e. a gain from padless to basal pads or vise versa a loss from basal to padless)
#We're actually only measuring the morphological differences and if the differences are consistent or 
#idiosyncratic across families 

#Now actual trajectory analyses
#Compare magnitude and angle of 
#padless vs basal 
#padless vs distal
#padless vs incipient
#Across families 
#Then do
#Within family comparisons to ask how similar padless-incipiant and padless basal patterns are 

#In this section, we do 50 comparisons 
#12 in padless vs basal
#12 in padless vs distal
#6 in padless vs incipient
#6 in Diplo comparisons
#6 in Sphaero comparisons
#2 in Phyllo comparisons
#6 in Gekkonidae comparisons

#Will correct for multiple comparisons using false discovery rate (p.adjust()) assuming 50 comparisons


#Padless vs basal
cats<-catassign[catassign$category=="padless" | catassign$category=="basal_pad",]
Coords<-dat.super$coords[,,catassign$category=="padless" | catassign$category=="basal_pad"]

#cut specimens from families that don't have both padless and basal pads
keep<-cats$Family!="Spheno"  #gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

keep<-cats$Family!="Eublepharidae"  #gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

keep<-cats$Family!="Carpho"  #gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]


rownames(cats)<-cats$species
rownames(cats)==dimnames(Coords)[[3]]
#should all be true

rm(keep)

#make a new object (it will be a "list") that contains our coord data, our clade assignments, and within group assignments ie pad class
#made composite list that includes coords, clade, and padclass
traj.dat<-list(coords=two.d.array(Coords), fam=as.factor(cats$Family), pad=as.factor(cats$category))

fit <- lm.rrpp(coords ~ fam * pad, data = traj.dat, iter = 10000) #should be 10000?
#reveal.model.designs(fit)
TA <- trajectory.analysis(fit, groups = as.factor(cats$Family), traj.pts = as.factor(cats$category), print.progress = TRUE)

rm(fit)

# Magnitude difference (absolute difference between path distances)
summary(TA, attribute = "MD") 
#Observed path distances by group

#.   Diplo     Gekko    Phyllo   Sphaero 
#0.7662249 0.8234191 0.9977553 0.7185162 

#Pairwise absolute differences in path distances, plus statistics
#                       d  UCL (95%)          Z     Pr > d
#Diplo:Gekko    0.05719421 0.15458502  0.2279174 0.42465753
#Diplo:Phyllo   0.23153038 0.16094592  2.2323855 0.01199880 *
#Diplo:Sphaero  0.04770872 0.17500124 -0.1358592 0.56564344
#Gekko:Phyllo   0.17433618 0.08973103  2.9711309 0.00009999 *
#Gekko:Sphaero  0.10490293 0.10869106  1.5058028 0.05859414 .
#Phyllo:Sphaero 0.27923910 0.12000352  3.3662671 0.00019998 *

#The only ones of significantly different length was phyllo vs everyone else.
#meaning padless vs basal in Phyllo was more different than padless vs basal in any other family. 

Mags<-summary(TA, attribute = "MD")$summary.table[,"Pr > d"]
names(Mags)<-rownames(summary(TA, attribute = "MD")$summary.table)
#Adjust the Pvals as needed
Mags<-p.adjust(p =  Mags, method = "fdr", n = 50)
Mags
#Diplo:Gekko   Diplo:Phyllo  Diplo:Sphaero   Gekko:Phyllo  Gekko:Sphaero Phyllo:Sphaero 
#1.0000000      0.1999800      1.0000000      0.0049995*      0.7324268      0.0049995* 
#Phyllo vs Sphaero and vs Gekko still different mags, but not vs Diplo


# Correlations (angles) between trajectories
summary(TA, attribute = "TC", angle.type = "deg") 

#Pairwise correlations between trajectories, plus statistics
#                      r     angle UCL (95%)        Z Pr > angle
#Diplo:Gekko    0.9717210 13.658366 12.336268 1.894103 0.02889711 *
#Diplo:Phyllo   0.9783730 11.937737 12.972869 1.369020 0.08299170 .
#Diplo:Sphaero  0.9673421 14.683206 13.728375 1.803889 0.03309669 *
#Gekko:Phyllo   0.9914038  7.518022  6.954498 1.913597 0.02769723 *
#Gekko:Sphaero  0.9913953  7.521705  8.393012 1.282900 0.09969003 .
#Phyllo:Sphaero 0.9820894 10.860330  9.352124 2.155873 0.01409859 *


Angs<-summary(TA, attribute = "TC", angle.type = "deg")$summary.table[,"Pr > angle"]
names(Angs)<-rownames(summary(TA, attribute = "TC", angle.type = "deg")$summary.table)
#Adjust the Pvals as needed
Angs<-p.adjust(p =  Angs, method = "fdr", n = 50)
Angs
#Diplo:Gekko   Diplo:Phyllo  Diplo:Sphaero   Gekko:Phyllo  Gekko:Sphaero Phyllo:Sphaero 
#0.4137086      0.8299170      0.4137086      0.4137086      0.8307503      0.4137086 
#none significantly different 


# Plot results just for confirmation. Will actually illustrate results on larger bgPCA plot with all the data on it
TP <- plot(TA, pch = as.numeric(traj.dat$fam) + 20, bg = as.numeric(traj.dat$pad), cex = 0.7, col = "gray")
add.trajectories(TP, traj.pch = unique(as.numeric(traj.dat$fam) + 20), start.bg = 1, end.bg = 2)
legend("topright", levels(traj.dat$fam), pch =  unique(as.numeric(traj.dat$fam) + 20), bty="n")
#text(rownames(TA$pca$x), x=TP$pc.points[,"PC1"], y=TP$pc.points[,"PC2"], cex=0.2)


#make actual large bgPCA plot to highlight specific comparisons 
#This same basic plot will be used for all the trajectory results plots 

#make a list of homogenized groups (family+pad class)
grps<-paste(catassign$Family, catassign$category) 

gPCA_out<-groupPCA(dat.super$coords, groups=grps) #save for all trajectory plots 
#this is slow

cols<-catassign$category #save for all trajectory plots 
cols[cols=="distal_pad"]<-"orange"
cols[cols=="padless"]<-"red"
cols[cols=="basal_pad"]<-"blue"
cols[cols=="incipient_pad"]<-"green"

pchs<-catassign$Family #save for all trajectory plots 
pchs[pchs=="Gekko"]<-0
pchs[pchs=="Eublepharidae"]<-5
pchs[pchs=="Sphaero"]<-2
pchs[pchs=="Carpho"]<-3
pchs[pchs=="Phyllo"]<-1
pchs[pchs=="Spheno"]<-4
pchs[pchs=="Diplo"]<-6
pchs<-as.numeric(pchs)

pdf(file=file.path("trajectory/padless-basal PC12.pdf"), width=7, height=8)

plot(gPCA_out$Scores[,1], pch=pchs, gPCA_out$Scores[,2], col=alpha(cols, 0.2), ylim=c(max(gPCA_out$Scores[,2])*1.8, min(gPCA_out$Scores[,2])), xlab=NA, ylab=NA, xaxt="n", yaxt="n") #ylim has max the min to reverse the y axis to make it similar to the true PCA plot. Also the lower value (technically the max) is *1/5 to give extra room at the bottom of the plot
mtext("Between Group Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Between Group Principal Component 2", side=2, line=0.2, cex=1.5)
title(main="Padless vs Basal Pads Across Families", line = 0.8, cex.main=1.5)

nams<-c("Sphenodon", "Carphodactylidae", "Diplodactylidae", "Eublepharidae", "Sphaerodactylidae", "Phyllodactylidae", "Gekkonidae")
legend("topleft", pch = c(4,3,6,5,2,1,0), legend = nams, bty = "n", cex=0.7)
legend("topleft", inset=c(0.2,0), legend=c("Padless", "Incipient pad", "Basal pad", "Distal pad"), text.col=c("red", "green", "blue" , "orange"), bty = "n", cex=0.7)

#highlight the points this analysis is focused on
out<-gPCA_out$Scores[rownames(gPCA_out$Scores) %in% rownames(cats),1:2]
points(out, pch=pchs[rownames(gPCA_out$Scores) %in% rownames(cats)], col=alpha(cols[rownames(gPCA_out$Scores) %in% rownames(cats)], 0.6))
rm(out)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko basal_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko basal_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko basal_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko basal_pad"],2][hpts], col="blue", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero basal_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero basal_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero basal_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero basal_pad"],2][hpts], col="blue", lwd=1.5)


#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo basal_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo basal_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo basal_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo basal_pad"],2][hpts], col="blue", lwd=1.5)


#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo basal_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo basal_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo basal_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo basal_pad"],2][hpts], col="blue", lwd=1.5)


segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko basal_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko basal_pad"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero basal_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero basal_pad"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo basal_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo basal_pad"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo basal_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo basal_pad"],2]))


#mean points
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2]), pch=22, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko basal_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko basal_pad"],2]), pch=22, col="black", bg="blue", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2]), pch=24, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero basal_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero basal_pad"],2]), pch=24, col="black", bg="blue", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],2]), pch=21, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo basal_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo basal_pad"],2]), pch=21, col="black", bg="blue", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2]), pch=25, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo basal_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo basal_pad"],2]), pch=25, col="black", bg="blue", cex=1.2)


#P values 
#magnitudes
mag<-as.expression(bquote(bold(Magnitudes)))

#function to display pvals
pvals_display<-function(p){
  #p<-0.07
  if(p>0.05){return(bquote(q~"="~.(as.character(round(p, 3)))))}
  if(p<=0.05 & p>=0.001){return(bquote(~bold(q~"="~.(as.character(round(p,3))))))}
  if(p<0.001){return(bquote(~bold(q~"<"~.(as.character(0.001)))))}
}


DvG<-as.expression(bquote(Diplodactylidae~vs~Gekkonidae*":"~.(pvals_display(Mags["Diplo:Gekko"]))))
DvP<-as.expression(bquote(Diplodactylidae~vs~Phyllodactylidae*":"~.(pvals_display(Mags["Diplo:Phyllo"]))))
DvS<-as.expression(bquote(Diplodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Mags["Diplo:Sphaero"]))))
GvP<-as.expression(bquote(Gekkonidae~vs~Phyllodactylidae*":"~.(pvals_display(Mags["Gekko:Phyllo"]))))
GvS<-as.expression(bquote(Gekkonidae~vs~Sphaerodactylidae*":"~.(pvals_display(Mags["Gekko:Sphaero"]))))
PvS<-as.expression(bquote(Phyllodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Mags["Phyllo:Sphaero"]))))


legend("bottomleft", legend = c(mag, DvG, DvP, DvS, GvP, GvS, PvS), bty = "n", cex=0.7, inset = c(-0.01,0))



#angles 
mag<-as.expression(bquote(bold(Angles)))

DvG<-as.expression(bquote(Diplodactylidae~vs~Gekkonidae*":"~.(pvals_display(Angs["Diplo:Gekko"]))))
DvP<-as.expression(bquote(Diplodactylidae~vs~Phyllodactylidae*":"~.(pvals_display(Angs["Diplo:Phyllo"]))))
DvS<-as.expression(bquote(Diplodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Angs["Diplo:Sphaero"]))))
GvP<-as.expression(bquote(Gekkonidae~vs~Phyllodactylidae*":"~.(pvals_display(Angs["Gekko:Phyllo"]))))
GvS<-as.expression(bquote(Gekkonidae~vs~Sphaerodactylidae*":"~.(pvals_display(Angs["Gekko:Sphaero"]))))
PvS<-as.expression(bquote(Phyllodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Angs["Phyllo:Sphaero"]))))


legend("bottomright", legend = c(mag, DvG, DvP, DvS, GvP, GvS, PvS), bty = "n", cex=0.7)


dev.off()



## visualize projected shape for the min and max of the bgPC axes. 
#bgPC1
proj.shape.2sd<- restoreShapes(1*sd(gPCA_out$Scores[,1]), gPCA_out$groupPCs[,1], gPCA_out$Grandmean)
proj.shape.2sd.neg<- restoreShapes(-1*sd(gPCA_out$Scores[,1]), gPCA_out$groupPCs[,1], gPCA_out$Grandmean)

proj.shape.2sd<-rotate.coords(proj.shape.2sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.2sd.neg<-rotate.coords(proj.shape.2sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("trajectory/bgPC1 shape 2sd.pdf"))
plot(proj.shape.2sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.2sd[,1],proj.shape.2sd[,2])
mtext(side=1, expression("" %->% "PC1 Maximum"), cex=2)
dev.off()

pdf(file=file.path("trajectory/bgPC1 shape -2sd.pdf"))
plot(proj.shape.2sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.2sd.neg[,1],proj.shape.2sd.neg[,2])
mtext(side=1, expression("PC1 Minimum" %<-% ""), cex=2)
dev.off()



#bgPC2
proj.shape.2sd<- restoreShapes(1*sd(gPCA_out$Scores[,2]), gPCA_out$groupPCs[,2], gPCA_out$Grandmean)
proj.shape.2sd.neg<- restoreShapes(-1*sd(gPCA_out$Scores[,2]), gPCA_out$groupPCs[,2], gPCA_out$Grandmean)

proj.shape.2sd<-rotate.coords(proj.shape.2sd, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise
proj.shape.2sd.neg<-rotate.coords(proj.shape.2sd.neg, type = "rotateCC") # the argument rotateCC rotates each matrix 90 deg counter clockwise

pdf(file=file.path("trajectory/bgPC2 shape 2sd.pdf"))
plot(proj.shape.2sd, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.2sd[,1],proj.shape.2sd[,2])
mtext(side=1,  expression("" %down% "PC2 Minimum"), cex=2) #remember the yaxis is inverted on this plot
dev.off()

pdf(file=file.path("trajectory/bgPC2 shape -2sd.pdf"))
plot(proj.shape.2sd.neg, asp=1, pch=19, xaxt="n", xlab=NA, yaxt="n", ylab=NA, col="white", bty="n")
GM_shape_plot(proj.shape.2sd.neg[,1],proj.shape.2sd.neg[,2])
mtext(side=1,  expression("" %up% "PC2 Maximum"), cex=2) #remember the yaxis is inverted on this plot
dev.off()



rm(Coords)
rm(cats)
rm(grps)
rm(proj.shape.2sd)
rm(proj.shape.2sd.neg)
rm(TA)
rm(TP)
rm(traj.dat)
rm(GvP)
rm(GvS)
rm(hpts)
rm(mag)
rm(nams)
rm(PvS)
rm(DvS)
rm(DvP)
rm(DvG)
rm(Mags)
rm(Angs)




#Padless vs distal
cats<-catassign[catassign$category=="padless" | catassign$category=="distal_pad",]
Coords<-dat.super$coords[,,catassign$category=="padless" | catassign$category=="distal_pad"]

#cut specimens from families that don't have both padless and distal pads
keep<-cats$Family!="Spheno"  #gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

keep<-cats$Family!="Eublepharidae"  #gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

keep<-cats$Family!="Carpho"  #gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

rownames(cats)<-cats$species
rownames(cats)==dimnames(Coords)[[3]]
#should all be true

rm(keep)

#make a new object (it will be a "list") that contains our coord data, our clade assignments, and within group assignments ie pad class
#made composite list that includes coords, clade, and padclass
traj.dat<-list(coords=two.d.array(Coords), fam=as.factor(cats$Family), pad=as.factor(cats$category))

fit <- lm.rrpp(coords ~ fam * pad, data = traj.dat, iter = 10000) #should be 10000?
#reveal.model.designs(fit)
TA <- trajectory.analysis(fit, groups = as.factor(cats$Family), traj.pts = as.factor(cats$category), print.progress = TRUE)

rm(fit)

# Magnitude difference (absolute difference between path distances)
summary(TA, attribute = "MD") 
#Observed path distances by group

#.   Diplo     Gekko    Phyllo   Sphaero 
#0.7286835 0.7774663 0.9858000 0.7198757 

#Pairwise absolute differences in path distances, plus statistics
#                        d UCL (95%)          Z     Pr > d
#Diplo:Gekko    0.048782768 0.1900026 -0.1736290 0.57844216
#Diplo:Phyllo   0.257116502 0.2019896  2.0368044 0.01859814 *
#Diplo:Sphaero  0.008807813 0.2028794 -1.4886808 0.92350765
#Gekko:Phyllo   0.208333734 0.1336955  2.5112012 0.00099990 *
#Gekko:Sphaero  0.057590580 0.1285514  0.3311921 0.38706129
#Phyllo:Sphaero 0.265924314 0.1445271  2.8581646 0.00049995 *

#The only ones of significantly different length was phyllo vs everyone else.
#meaning padless vs basal in Phyllo was more different than padless vs basal in any other family. 

Mags<-summary(TA, attribute = "MD")$summary.table[,"Pr > d"]
names(Mags)<-rownames(summary(TA, attribute = "MD")$summary.table)
#Adjust the Pvals as needed
Mags<-p.adjust(p =  Mags, method = "fdr", n = 50)
Mags
#Diplo:Gekko   Diplo:Phyllo  Diplo:Sphaero   Gekko:Phyllo  Gekko:Sphaero Phyllo:Sphaero 
#1.0000000      0.3099690      1.0000000      0.0249975*      1.0000000      0.0249975*
#Phyllo vs Gekko and vs Shpaero still sig


# Correlations (angles) between trajectories
summary(TA, attribute = "TC", angle.type = "deg") 

#Pairwise correlations between trajectories, plus statistics
#                       r    angle UCL (95%)        Z Pr > angle
#Diplo:Gekko    0.9405826 19.85037  15.57518 2.179562 0.01729827 *
#Diplo:Phyllo   0.8921536 26.85487  16.38153 2.807201 0.00319968 *
#Diplo:Sphaero  0.9268872 22.04533  16.34239 2.342105 0.01069893 *
#Gekko:Phyllo   0.9738124 13.14128  10.31249 2.320149 0.00899910 *
#Gekko:Sphaero  0.9624474 15.75165  10.16512 2.919348 0.00119988 *
#Phyllo:Sphaero 0.9624721 15.74644  11.37884 2.582059 0.00379962 *

Angs<-summary(TA, attribute = "TC", angle.type = "deg")$summary.table[,"Pr > angle"]
names(Angs)<-rownames(summary(TA, attribute = "TC", angle.type = "deg")$summary.table)
#Adjust the Pvals as needed
Angs<-p.adjust(p =  Angs, method = "fdr", n = 50)
Angs
#Diplo:Gekko   Diplo:Phyllo  Diplo:Sphaero   Gekko:Phyllo  Gekko:Sphaero Phyllo:Sphaero 
#0.1441523      0.0633270      0.1069893      0.1069893      0.0599940      0.0633270 
#none of the comparisons are sig anymore 


# Plot results just for confirmation. Will actually illustrate results on larger bgPCA plot with all the data on it
TP <- plot(TA, pch = as.numeric(traj.dat$fam) + 20, bg = as.numeric(traj.dat$pad), cex = 0.7, col = "gray")
add.trajectories(TP, traj.pch = unique(as.numeric(traj.dat$fam) + 20), start.bg = 1, end.bg = 2)
legend("topright", levels(traj.dat$fam), pch =  unique(as.numeric(traj.dat$fam) + 20), bty="n")
#text(rownames(TA$pca$x), x=TP$pc.points[,"PC1"], y=TP$pc.points[,"PC2"], cex=0.2)


#make actual large bgPCA plot to highlight specific comparisons 

pdf(file=file.path("trajectory/padless-distal PC12.pdf"), width=7, height=8)

plot(gPCA_out$Scores[,1], pch=pchs, gPCA_out$Scores[,2], col=alpha(cols, 0.2), ylim=c(max(gPCA_out$Scores[,2])*1.8, min(gPCA_out$Scores[,2])), xlab=NA, ylab=NA, xaxt="n", yaxt="n") #ylim has max the min to reverse the y axis to make it similar to the true PCA plot. Also the lower value (technically the max) is *1/5 to give extra room at the bottom of the plot
mtext("Between Group Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Between Group Principal Component 2", side=2, line=0.2, cex=1.5)
title(main="Padless vs Distal Pads Across Families", line = 0.8, cex.main=1.5)

#highlight the points this analysis is focused on
out<-gPCA_out$Scores[rownames(gPCA_out$Scores) %in% rownames(cats),1:2]
points(out, pch=pchs[rownames(gPCA_out$Scores) %in% rownames(cats)], col=alpha(cols[rownames(gPCA_out$Scores) %in% rownames(cats)], 0.6))
rm(out)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko distal_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko distal_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko distal_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko distal_pad"],2][hpts], col="orange", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero distal_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero distal_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero distal_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero distal_pad"],2][hpts], col="orange", lwd=1.5)


#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo distal_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo distal_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo distal_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo distal_pad"],2][hpts], col="orange", lwd=1.5)


#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo distal_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo distal_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo distal_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo distal_pad"],2][hpts], col="orange", lwd=1.5)


segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko distal_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko distal_pad"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero distal_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero distal_pad"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo distal_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo distal_pad"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo distal_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo distal_pad"],2]))


#mean points
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2]), pch=22, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko distal_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko distal_pad"],2]), pch=22, col="black", bg="orange", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2]), pch=24, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero distal_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero distal_pad"],2]), pch=24, col="black", bg="orange", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo padless"],2]), pch=21, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo distal_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Phyllo distal_pad"],2]), pch=21, col="black", bg="orange", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2]), pch=25, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo distal_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo distal_pad"],2]), pch=25, col="black", bg="orange", cex=1.2)



#P values 
#magnitudes
mag<-as.expression(bquote(bold(Magnitudes)))

DvG<-as.expression(bquote(Diplodactylidae~vs~Gekkonidae*":"~.(pvals_display(Mags["Diplo:Gekko"]))))
DvP<-as.expression(bquote(Diplodactylidae~vs~Phyllodactylidae*":"~.(pvals_display(Mags["Diplo:Phyllo"]))))
DvS<-as.expression(bquote(Diplodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Mags["Diplo:Sphaero"]))))
GvP<-as.expression(bquote(Gekkonidae~vs~Phyllodactylidae*":"~.(pvals_display(Mags["Gekko:Phyllo"]))))
GvS<-as.expression(bquote(Gekkonidae~vs~Sphaerodactylidae*":"~.(pvals_display(Mags["Gekko:Sphaero"]))))
PvS<-as.expression(bquote(Phyllodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Mags["Phyllo:Sphaero"]))))


legend("bottomleft", legend = c(mag, DvG, DvP, DvS, GvP, GvS, PvS), bty = "n", cex=0.7, inset = c(-0.01,0))



#angles 
mag<-as.expression(bquote(bold(Angles)))

DvG<-as.expression(bquote(Diplodactylidae~vs~Gekkonidae*":"~.(pvals_display(Angs["Diplo:Gekko"]))))
DvP<-as.expression(bquote(Diplodactylidae~vs~Phyllodactylidae*":"~.(pvals_display(Angs["Diplo:Phyllo"]))))
DvS<-as.expression(bquote(Diplodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Angs["Diplo:Sphaero"]))))
GvP<-as.expression(bquote(Gekkonidae~vs~Phyllodactylidae*":"~.(pvals_display(Angs["Gekko:Phyllo"]))))
GvS<-as.expression(bquote(Gekkonidae~vs~Sphaerodactylidae*":"~.(pvals_display(Angs["Gekko:Sphaero"]))))
PvS<-as.expression(bquote(Phyllodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Angs["Phyllo:Sphaero"]))))


legend("bottomright", legend = c(mag, DvG, DvP, DvS, GvP, GvS, PvS), bty = "n", cex=0.7)


dev.off()







rm(Coords)
rm(cats)
rm(TA)
rm(TP)
rm(traj.dat)
rm(GvP)
rm(GvS)
rm(hpts)
rm(mag)
rm(PvS)
rm(DvS)
rm(DvG)
rm(DvP)
rm(Angs)
rm(Mags)


#Padless vs incipient
cats<-catassign[catassign$category=="padless" | catassign$category=="incipient_pad",]
Coords<-dat.super$coords[,,catassign$category=="padless" | catassign$category=="incipient_pad"]

#cut specimens from families that don't have both padless and incipient pad
keep<-cats$Family!="Spheno"  #only padless, gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

keep<-cats$Family!="Eublepharidae" #only padless, gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

keep<-cats$Family!="Carpho" #only padless, gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

keep<-cats$Family!="Phyllo" #doesn't have any incipient pads, gotta use an intermediate variable 
cats<-cats[keep,]
Coords<-Coords[,,keep]

rownames(cats)<-cats$species
rownames(cats)==dimnames(Coords)[[3]]
#should all be true

rm(keep)

#make a new object (it will be a "list") that contains our coord data, our clade assignments, and within group assignments ie pad class
#made composite list that includes coords, clade, and padclass
traj.dat<-list(coords=two.d.array(Coords), fam=as.factor(cats$Family), pad=as.factor(cats$category))

fit <- lm.rrpp(coords ~ fam * pad, data = traj.dat, iter = 10000) #should be 10000?
#reveal.model.designs(fit)
TA <- trajectory.analysis(fit, groups = as.factor(cats$Family), traj.pts = as.factor(cats$category), print.progress = TRUE)

rm(fit)

# Magnitude difference (absolute difference between path distances)
summary(TA, attribute = "MD") 
#Observed path distances by group

#Diplo     Gekko   Sphaero 
#0.3078917 0.2024415 0.2049582 

#Pairwise absolute differences in path distances, plus statistics
#                        d  UCL (95%)          Z    Pr > d
#Diplo:Gekko   0.105450207 0.15886780  0.9951923 0.1685831
#Diplo:Sphaero 0.102933479 0.15731704  0.9653779 0.1718828
#Gekko:Sphaero 0.002516728 0.08525406 -1.7578831 0.9556044

#none are different
Mags<-summary(TA, attribute = "MD")$summary.table[,"Pr > d"]
names(Mags)<-rownames(summary(TA, attribute = "MD")$summary.table)
#Adjust the Pvals as needed
Mags<-p.adjust(p =  Mags, method = "fdr", n = 50)
Mags
#Diplo:Gekko Diplo:Sphaero Gekko:Sphaero 
#1             1             1 



# Correlations (angles) between trajectories
summary(TA, attribute = "TC", angle.type = "deg") 

#Pairwise correlations between trajectories, plus statistics
#                      r    angle UCL (95%)        Z Pr > angle
#Diplo:Gekko   0.5518738 56.50434  67.18575 1.281579 0.09459054
#Diplo:Sphaero 0.5374429 57.49027  69.09088 1.229573 0.11098890
#Gekko:Sphaero 0.7690193 39.73410  41.60673 1.555670 0.06319368

Angs<-summary(TA, attribute = "TC", angle.type = "deg")$summary.table[,"Pr > angle"]
names(Angs)<-rownames(summary(TA, attribute = "TC", angle.type = "deg")$summary.table)
#Adjust the Pvals as needed
Angs<-p.adjust(p =  Angs, method = "fdr", n = 50)
Angs
#Diplo:Gekko Diplo:Sphaero Gekko:Sphaero 
#1             1             1 



# Plot results just for confirmation. Will actually illustrate results on larger bgPCA plot with all the data on it
TP <- plot(TA, pch = as.numeric(traj.dat$fam) + 20, bg = as.numeric(traj.dat$pad), cex = 0.7, col = "gray")
add.trajectories(TP, traj.pch = unique(as.numeric(traj.dat$fam) + 20), start.bg = 1, end.bg = 2)
legend("topright", levels(traj.dat$fam), pch =  unique(as.numeric(traj.dat$fam) + 20), bty="n")
#text(rownames(TA$pca$x), x=TP$pc.points[,"PC1"], y=TP$pc.points[,"PC2"], cex=0.2)


#make actual large bgPCA plot to highlight specific comparisons 

pdf(file=file.path("trajectory/padless-incipient PC12.pdf"), width=7, height=8)

plot(gPCA_out$Scores[,1], pch=pchs, gPCA_out$Scores[,2], col=alpha(cols, 0.2), ylim=c(max(gPCA_out$Scores[,2])*1.8, min(gPCA_out$Scores[,2])), xlab=NA, ylab=NA, xaxt="n", yaxt="n") #ylim has max the min to reverse the y axis to make it similar to the true PCA plot. Also the lower value (technically the max) is *1/5 to give extra room at the bottom of the plot
mtext("Between Group Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Between Group Principal Component 2", side=2, line=0.2, cex=1.5)
title(main="Padless vs Incipient Pads Across Families", line = 0.8, cex.main=1.5)

#highlight the points this analysis is focused on
out<-gPCA_out$Scores[rownames(gPCA_out$Scores) %in% rownames(cats),1:2]
points(out, pch=pchs[rownames(gPCA_out$Scores) %in% rownames(cats)], col=alpha(cols[rownames(gPCA_out$Scores) %in% rownames(cats)], 0.6))
rm(out)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko incipient_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko incipient_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko incipient_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko incipient_pad"],2][hpts], col="green", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero incipient_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero incipient_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero incipient_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero incipient_pad"],2][hpts], col="green", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2][hpts], col="red", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo incipient_pad"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo incipient_pad"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo incipient_pad"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo incipient_pad"],2][hpts], col="green", lwd=1.5)


segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko incipient_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko incipient_pad"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero incipient_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero incipient_pad"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo incipient_pad"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo incipient_pad"],2]))


#mean points
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko padless"],2]), pch=22, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko incipient_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Gekko incipient_pad"],2]), pch=22, col="black", bg="green", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero padless"],2]), pch=24, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero incipient_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Sphaero incipient_pad"],2]), pch=24, col="black", bg="green", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo padless"],2]), pch=25, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo incipient_pad"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$category)=="Diplo incipient_pad"],2]), pch=25, col="black", bg="green", cex=1.2)



#P values 
#magnitudes
#P values 
#magnitudes
mag<-as.expression(bquote(bold(Magnitudes)))

DvG<-as.expression(bquote(Diplodactylidae~vs~Gekkonidae*":"~.(pvals_display(Mags["Diplo:Gekko"]))))
DvS<-as.expression(bquote(Diplodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Mags["Diplo:Sphaero"]))))
GvS<-as.expression(bquote(Gekkonidae~vs~Sphaerodactylidae*":"~.(pvals_display(Mags["Gekko:Sphaero"]))))

legend("bottomleft", legend = c(mag, DvG, DvS, GvS), bty = "n", cex=0.7, inset = c(-0.01,0))



#angles 
mag<-as.expression(bquote(bold(Angles)))

DvG<-as.expression(bquote(Diplodactylidae~vs~Gekkonidae*":"~.(pvals_display(Angs["Diplo:Gekko"]))))
DvS<-as.expression(bquote(Diplodactylidae~vs~Sphaerodactylidae*":"~.(pvals_display(Angs["Diplo:Sphaero"]))))
GvS<-as.expression(bquote(Gekkonidae~vs~Sphaerodactylidae*":"~.(pvals_display(Angs["Gekko:Sphaero"]))))


legend("bottomright", legend = c(mag, DvG, DvS, GvS), bty = "n", cex=0.7)


dev.off()



rm(Coords)
rm(cats)
rm(TA)
rm(TP)
rm(traj.dat)
rm(GvS)
rm(hpts)
rm(mag)
rm(DvG)
rm(DvS)
rm(Mags)
rm(Angs)






#Within Gekko comparisons
cats<-catassign[catassign$Family=="Gekko",]
Coords<-dat.super$coords[,,catassign$Family=="Gekko"]


#have to duplicate padless specimens so that multiple comparisons can be made
#basal will pair with padless under the basal-family
#incipient will pair with padless2 under the inc-family
#distal will pair with padless3 under the distal-family
#we'll compare padless vs padded within each of our mock families


toadd<-cats[which(cats$category=="padless"),]
toadd$Family<-"inc-family"
toadd$category<-"padless2"
cats[cats$category=="padless","Family"]<-"basal-family"
cats<-rbind(cats, toadd)

toadd$Family<-"distal-family"
toadd$category<-"padless3"
cats<-rbind(cats, toadd)

cats[cats$category=="incipient_pad","Family"]<-"inc-family"
cats[cats$category=="distal_pad","Family"]<-"distal-family"
cats[cats$category=="basal_pad","Family"]<-"basal-family"


toadd<-cats$category
toadd[toadd=="padless2"]<-"padless"
toadd[toadd=="padless3"]<-"padless"
toadd[toadd=="basal_pad"]<-"padded"
toadd[toadd=="distal_pad"]<-"padded"
toadd[toadd=="incipient_pad"]<-"padded"

cats<-cbind(cats, toadd)
colnames(cats)[4]<-"general_padcat"

#unique(paste(cats$general_padcat, cats$Family))

toadd<-Coords[,,which(cats$category=="padless")]

NewCoords<-array(data = NA, dim = c(dim(Coords)[1], dim(Coords)[2],  dim(Coords)[3]+dim(toadd)[3]+dim(toadd)[3] ))
NewCoords[,,1:dim(Coords)[3]]<-Coords
NewCoords[,,(dim(Coords)[3]+1):(dim(Coords)[3]+dim(toadd)[3])]<-toadd
NewCoords[,,(dim(Coords)[3]+dim(toadd)[3]+1):(dim(Coords)[3]+dim(toadd)[3]+dim(toadd)[3])]<-toadd
dimnames(NewCoords)[[3]]<-rownames(cats)


rownames(cats)==dimnames(NewCoords)[[3]]
#should all be true

rm(toadd)
rm(Coords)

#make a new object (it will be a "list") that contains our coord data, our clade assignments, and within group assignments ie pad class
#made composite list that includes coords, clade, and padclass
traj.dat<-list(coords=two.d.array(NewCoords), pad=as.factor(cats$general_padcat), fam=as.factor(cats$Family))

fit <- lm.rrpp(coords ~ fam * pad, data = traj.dat, iter = 10000) #should be 10000?
#reveal.model.designs(fit)
TA <- trajectory.analysis(fit, groups = as.factor(cats$Family), traj.pts = as.factor(cats$general_padcat), print.progress = TRUE)

rm(fit)




# Magnitude difference (absolute difference between path distances)
summary(TA, attribute = "MD") 
#Observed path distances by group

#basal-family distal-family    inc-family 
#0.8234191     0.7774663     0.2024415 

#Pairwise absolute differences in path distances, plus statistics
#                                    d UCL (95%)          Z     Pr > d
#basal-family:distal-family 0.04595284 0.1704158 -0.2613151 0.60563944
#basal-family:inc-family    0.62097758 0.1589909  4.9010143 0.00009999 *
#distal-family:inc-family   0.57502473 0.1883985  4.1888658 0.00009999 *

#padless to incipient is different mag than all the other padless vs padded comparisons
Mags<-summary(TA, attribute = "MD")$summary.table[,"Pr > d"]
names(Mags)<-rownames(summary(TA, attribute = "MD")$summary.table)
#Adjust the Pvals as needed
Mags<-p.adjust(p =  Mags, method = "fdr", n = 50)
Mags
#basal-family:distal-family    basal-family:inc-family   distal-family:inc-family 
#1.00000000                 0.00249975                 0.00249975
#same results 


# Correlations (angles) between trajectories
summary(TA, attribute = "TC", angle.type = "deg") 

#Pairwise correlations between trajectories, plus statistics
#                                   r    angle UCL (95%)        Z Pr > angle
#basal-family:distal-family 0.9068912 24.92079  11.72536 3.715942  9.999e-05 *
#basal-family:inc-family    0.4578433 62.75197  11.00459 6.479962  9.999e-05 *
#distal-family:inc-family   0.7458046 41.77175  13.13713 4.761582  9.999e-05 *

Angs<-summary(TA, attribute = "TC", angle.type = "deg")$summary.table[,"Pr > angle"]
names(Angs)<-rownames(summary(TA, attribute = "TC", angle.type = "deg")$summary.table)
#Adjust the Pvals as needed
Angs<-p.adjust(p =  Angs, method = "fdr", n = 50)
Angs
#basal-family:distal-family    basal-family:inc-family   distal-family:inc-family 
#0.0016665                  0.0016665                  0.0016665 
#same results 



# Plot results just for confirmation. Will actually illustrate results on larger bgPCA plot with all the data on it
TP <- plot(TA, pch = as.numeric(traj.dat$fam) + 20, bg = as.numeric(traj.dat$pad), cex = 0.7, col = "gray")
add.trajectories(TP, traj.pch = unique(as.numeric(traj.dat$fam) + 20), start.bg = 1, end.bg = 2)
legend("topright", levels(traj.dat$fam), pch =  unique(as.numeric(traj.dat$fam) + 20), bty="n")
#text(rownames(TA$pca$x), x=TP$pc.points[,"PC1"], y=TP$pc.points[,"PC2"], cex=0.2)


#make actual large bgPCA plot to highlight specific comparisons 

pdf(file=file.path("trajectory/Gekko PC12.pdf"), width=7, height=7)

plot(gPCA_out$Scores[,1], pch=pchs, gPCA_out$Scores[,2], col=alpha(cols, 0.2), ylim=c(max(gPCA_out$Scores[,2])*1.5, min(gPCA_out$Scores[,2])), xlab=NA, ylab=NA, xaxt="n", yaxt="n") #ylim has max the min to reverse the y axis to make it similar to the true PCA plot. Also the lower value (technically the max) is *1/5 to give extra room at the bottom of the plot
mtext("Between Group Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Between Group Principal Component 2", side=2, line=0.2, cex=1.5)
title(main="Gekkonidae Toe Pad Class Comparisons", line = 0.8, cex.main=1.5)

#highlight the points this analysis is focused on
out<-gPCA_out$Scores[rownames(gPCA_out$Scores) %in% rownames(cats),1:2]
points(out, pch=pchs[rownames(gPCA_out$Scores) %in% rownames(cats)], col=alpha(cols[rownames(gPCA_out$Scores) %in% rownames(cats)], 0.6))
rm(out)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],1], gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],1][hpts], gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],2][hpts], col="red", lwd=1.5)


#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2][hpts], col="blue", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2][hpts], col="orange", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2][hpts], col="green", lwd=1.5)



segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2]))




#mean points
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],2]), pch=22, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2]), pch=22, col="black", bg="green", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2]), pch=22, col="black", bg="orange", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2]), pch=22, col="black", bg="blue", cex=1.2)

#P values 
#magnitudes
mag<-as.expression(bquote(bold(Magnitudes)))

BvD<-as.expression(bquote(Padlass-Basal~vs~Padless-Distal*":"~.(pvals_display(Mags["basal-family:distal-family"]))))
BvI<-as.expression(bquote(Padlass-Basal~vs~Padless-Incipient*":"~.(pvals_display(Mags["basal-family:inc-family"]))))
DvI<-as.expression(bquote(Padlass-Distal~vs~Padless-Incipient*":"~.(pvals_display(Mags["distal-family:inc-family"]))))

legend("bottomleft", legend = c(mag, BvD, BvI, DvI), bty = "n", cex=0.7, inset = c(-0.01,0))



#angles 
mag<-as.expression(bquote(bold(Angles)))

BvD<-as.expression(bquote(Padlass-Basal~vs~Padless-Distal*":"~.(pvals_display(Angs["basal-family:distal-family"]))))
BvI<-as.expression(bquote(Padlass-Basal~vs~Padless-Incipient*":"~.(pvals_display(Angs["basal-family:inc-family"]))))
DvI<-as.expression(bquote(Padlass-Distal~vs~Padless-Incipient*":"~.(pvals_display(Angs["distal-family:inc-family"]))))


legend("bottomright", legend = c(mag, BvD, BvI, DvI), bty = "n", cex=0.7)


dev.off()



rm(NewCoords)
rm(cats)
rm(TA)
rm(TP)
rm(traj.dat)
rm(BvI)
rm(BvD)
rm(DvI)
rm(hpts)
rm(mag)
rm(Mags)
rm(Angs)



#Within Sphaero comparisons
cats<-catassign[catassign$Family=="Sphaero",]
Coords<-dat.super$coords[,,catassign$Family=="Sphaero"]


#have to duplicate padless specimens so that multiple comparisons can be made
#basal will pair with padless under the  basal-family
#incipient will pair with padless2 under the inc-family
#distal will pair with padless3 under the distal-family
#we'll compare padless vs padded within each of our mock families


toadd<-cats[which(cats$category=="padless"),]
toadd$Family<-"inc-family"
toadd$category<-"padless2"
cats[cats$category=="padless","Family"]<-"basal-family"
cats<-rbind(cats, toadd)

toadd$Family<-"distal-family"
toadd$category<-"padless3"
cats<-rbind(cats, toadd)

cats[cats$category=="incipient_pad","Family"]<-"inc-family"
cats[cats$category=="distal_pad","Family"]<-"distal-family"
cats[cats$category=="basal_pad","Family"]<-"basal-family"


toadd<-cats$category
toadd[toadd=="padless2"]<-"padless"
toadd[toadd=="padless3"]<-"padless"
toadd[toadd=="basal_pad"]<-"padded"
toadd[toadd=="distal_pad"]<-"padded"
toadd[toadd=="incipient_pad"]<-"padded"

cats<-cbind(cats, toadd)
colnames(cats)[4]<-"general_padcat"

#unique(paste(cats$general_padcat, cats$Family))

toadd<-Coords[,,which(cats$category=="padless")]

NewCoords<-array(data = NA, dim = c(dim(Coords)[1], dim(Coords)[2],  dim(Coords)[3]+dim(toadd)[3]+dim(toadd)[3] ))
NewCoords[,,1:dim(Coords)[3]]<-Coords
NewCoords[,,(dim(Coords)[3]+1):(dim(Coords)[3]+dim(toadd)[3])]<-toadd
NewCoords[,,(dim(Coords)[3]+dim(toadd)[3]+1):(dim(Coords)[3]+dim(toadd)[3]+dim(toadd)[3])]<-toadd
dimnames(NewCoords)[[3]]<-rownames(cats)


rownames(cats)==dimnames(NewCoords)[[3]]
#should all be true

rm(toadd)
rm(Coords)

#make a new object (it will be a "list") that contains our coord data, our clade assignments, and within group assignments ie pad class
#made composite list that includes coords, clade, and padclass
traj.dat<-list(coords=two.d.array(NewCoords), pad=as.factor(cats$general_padcat), fam=as.factor(cats$Family))

fit <- lm.rrpp(coords ~ fam * pad, data = traj.dat, iter = 10000) #should be 10000?
reveal.model.designs(fit)
TA <- trajectory.analysis(fit, groups = as.factor(cats$Family), traj.pts = as.factor(cats$general_padcat), print.progress = TRUE)

rm(fit)




# Magnitude difference (absolute difference between path distances)
summary(TA, attribute = "MD") 
#Observed path distances by group

#basal-family distal-family    inc-family 
#0.7233174     0.7235215     0.2065929 

#Pairwise absolute differences in path distances, plus statistics
#                                    d UCL (95%)         Z     Pr > d
#basal-family:distal-family 0.001359506 0.1809292 -2.200050 0.99080092
#basal-family:inc-family    0.513557920 0.1718153  4.117850 0.00009999 *
#distal-family:inc-family   0.514917426 0.1690069  4.203546 0.00009999 *


#padless to incepient is different than all the other padless->padded comparisons
Mags<-summary(TA, attribute = "MD")$summary.table[,"Pr > d"]
names(Mags)<-rownames(summary(TA, attribute = "MD")$summary.table)
#Adjust the Pvals as needed
Mags<-p.adjust(p =  Mags, method = "fdr", n = 50)
Mags
#basal-family:distal-family    basal-family:inc-family   distal-family:inc-family 
#1.00000000                 0.00249975                 0.00249975 
#no change


# Correlations (angles) between trajectories
summary(TA, attribute = "TC", angle.type = "deg") 

#Pairwise correlations between trajectories, plus statistics
#                                   r    angle UCL (95%)        Z Pr > angle
#basal-family:distal-family 0.8756875 28.87355  16.88734 3.162004 0.00079992 *
#basal-family:inc-family    0.4939999 60.39618  16.24160 5.519434 0.00009999 *
#distal-family:inc-family   0.6399194 50.21419  15.66186 5.218305 0.00009999 *

#all different angles 
Angs<-summary(TA, attribute = "TC", angle.type = "deg")$summary.table[,"Pr > angle"]
names(Angs)<-rownames(summary(TA, attribute = "TC", angle.type = "deg")$summary.table)
#Adjust the Pvals as needed
Angs<-p.adjust(p =  Angs, method = "fdr", n = 50)
Angs
#basal-family:distal-family    basal-family:inc-family   distal-family:inc-family 
#0.01333200                 0.00249975                 0.00249975 
#no change


# Plot results just for confirmation. Will actually illustrate results on larger bgPCA plot with all the data on it
TP <- plot(TA, pch = as.numeric(traj.dat$fam) + 20, bg = as.numeric(traj.dat$pad), cex = 0.7, col = "gray")
add.trajectories(TP, traj.pch = unique(as.numeric(traj.dat$fam) + 20), start.bg = 1, end.bg = 2)
legend("topright", levels(traj.dat$fam), pch =  unique(as.numeric(traj.dat$fam) + 20), bty="n")
#text(rownames(TA$pca$x), x=TP$pc.points[,"PC1"], y=TP$pc.points[,"PC2"], cex=0.2)


#make actual large bgPCA plot to highlight specific comparisons 

pdf(file=file.path("trajectory/Sphaero PC12.pdf"), width=7, height=7)

plot(gPCA_out$Scores[,1], pch=pchs, gPCA_out$Scores[,2], col=alpha(cols, 0.2), ylim=c(max(gPCA_out$Scores[,2])*1.5, min(gPCA_out$Scores[,2])), xlab=NA, ylab=NA, xaxt="n", yaxt="n") #ylim has max the min to reverse the y axis to make it similar to the true PCA plot. Also the lower value (technically the max) is *1/5 to give extra room at the bottom of the plot
mtext("Between Group Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Between Group Principal Component 2", side=2, line=0.2, cex=1.5)
title(main="Sphaerodactylidae Toe Pad Class Comparisons", line = 0.8, cex.main=1.5)

#highlight the points this analysis is focused on
out<-gPCA_out$Scores[rownames(gPCA_out$Scores) %in% rownames(cats),1:2]
points(out, pch=pchs[rownames(gPCA_out$Scores) %in% rownames(cats)], col=alpha(cols[rownames(gPCA_out$Scores) %in% rownames(cats)], 0.6))
rm(out)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],1], gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],1][hpts], gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],2][hpts], col="red", lwd=1.5)


#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2][hpts], col="blue", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2][hpts], col="orange", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2][hpts], col="green", lwd=1.5)



segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2]))




#mean points
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],2]), pch=24, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2]), pch=24, col="black", bg="green", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2]), pch=24, col="black", bg="orange", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2]), pch=24, col="black", bg="blue", cex=1.2)

#P values 
#magnitudes
mag<-as.expression(bquote(bold(Magnitudes)))

BvD<-as.expression(bquote(Padlass-Basal~vs~Padless-Distal*":"~.(pvals_display(Mags["basal-family:distal-family"]))))
BvI<-as.expression(bquote(Padlass-Basal~vs~Padless-Incipient*":"~.(pvals_display(Mags["basal-family:inc-family"]))))
DvI<-as.expression(bquote(Padlass-Distal~vs~Padless-Incipient*":"~.(pvals_display(Mags["distal-family:inc-family"]))))

legend("bottomleft", legend = c(mag, BvD, BvI, DvI), bty = "n", cex=0.7, inset = c(-0.01,0))



#angles 
mag<-as.expression(bquote(bold(Angles)))

BvD<-as.expression(bquote(Padlass-Basal~vs~Padless-Distal*":"~.(pvals_display(Angs["basal-family:distal-family"]))))
BvI<-as.expression(bquote(Padlass-Basal~vs~Padless-Incipient*":"~.(pvals_display(Angs["basal-family:inc-family"]))))
DvI<-as.expression(bquote(Padlass-Distal~vs~Padless-Incipient*":"~.(pvals_display(Angs["distal-family:inc-family"]))))


legend("bottomright", legend = c(mag, BvD, BvI, DvI), bty = "n", cex=0.7)

dev.off()



rm(NewCoords)
rm(cats)
rm(TA)
rm(TP)
rm(traj.dat)
rm(BvI)
rm(BvD)
rm(DvI)
rm(hpts)
rm(mag)
rm(Mags)
rm(Angs)







#Within Phyllo comparisons
cats<-catassign[catassign$Family=="Phyllo",]
Coords<-dat.super$coords[,,catassign$Family=="Phyllo"]


#have to duplicate padless specimens so that multiple comparisons can be made
#basal will pair with padless under the  basal-family
#distal will pair with padless2 under the distal-family
#we'll compare padless vs padded within each of our mock families
#Phyllo doesnt have any incipient

toadd<-cats[which(cats$category=="padless"),]
toadd$Family<-"distal-family"
toadd$category<-"padless2"
cats[cats$category=="padless","Family"]<-"basal-family"
cats<-rbind(cats, toadd)

cats[cats$category=="distal_pad","Family"]<-"distal-family"
cats[cats$category=="basal_pad","Family"]<-"basal-family"


toadd<-cats$category
toadd[toadd=="padless2"]<-"padless"
toadd[toadd=="basal_pad"]<-"padded"
toadd[toadd=="distal_pad"]<-"padded"

cats<-cbind(cats, toadd)
colnames(cats)[4]<-"general_padcat"

#unique(paste(cats$general_padcat, cats$Family))

toadd<-Coords[,,which(cats$category=="padless")]

NewCoords<-array(data = NA, dim = c(dim(Coords)[1], dim(Coords)[2],  dim(Coords)[3]+dim(toadd)[3] ))
NewCoords[,,1:dim(Coords)[3]]<-Coords
NewCoords[,,(dim(Coords)[3]+1):(dim(Coords)[3]+dim(toadd)[3])]<-toadd
dimnames(NewCoords)[[3]]<-rownames(cats)


rownames(cats)==dimnames(NewCoords)[[3]]
#should all be true

rm(toadd)
rm(Coords)

#make a new object (it will be a "list") that contains our coord data, our clade assignments, and within group assignments ie pad class
#made composite list that includes coords, clade, and padclass
traj.dat<-list(coords=two.d.array(NewCoords), pad=as.factor(cats$general_padcat), fam=as.factor(cats$Family))

fit <- lm.rrpp(coords ~ fam * pad, data = traj.dat, iter = 10000) #should be 10000?
#reveal.model.designs(fit)
TA <- trajectory.analysis(fit, groups = as.factor(cats$Family), traj.pts = as.factor(cats$general_padcat), print.progress = TRUE)

rm(fit)




# Magnitude difference (absolute difference between path distances)
summary(TA, attribute = "MD") 
#Observed path distances by group

#basal-family distal-family 
#0.9977553     0.9858000 

#Pairwise absolute differences in path distances, plus statistics
#                                    d  UCL (95%)           Z    Pr > d
#basal-family:distal-family 0.01195528 0.06659217 -0.6180954 0.7180282

#not different 
Mags<-summary(TA, attribute = "MD")$summary.table[,"Pr > d"]
names(Mags)<-rownames(summary(TA, attribute = "MD")$summary.table)
#Adjust the Pvals as needed
Mags<-p.adjust(p =  Mags, method = "fdr", n = 50)
Mags
#basal-family:distal-family 
#1 


# Correlations (angles) between trajectories
summary(TA, attribute = "TC", angle.type = "deg") 

#Pairwise correlations between trajectories, plus statistics
#                                   r    angle UCL (95%)        Z Pr > angle
#basal-family:distal-family 0.9071164 24.89015  7.696471 3.897913  9.999e-05 *

# different angles
Angs<-summary(TA, attribute = "TC", angle.type = "deg")$summary.table[,"Pr > angle"]
names(Angs)<-rownames(summary(TA, attribute = "TC", angle.type = "deg")$summary.table)
#Adjust the Pvals as needed
Angs<-p.adjust(p =  Angs, method = "fdr", n = 50)
Angs
#basal-family:distal-family 
#0.0049995 



# Plot results just for confirmation. Will actually illustrate results on larger bgPCA plot with all the data on it
TP <- plot(TA, pch = as.numeric(traj.dat$fam) + 20, bg = as.numeric(traj.dat$pad), cex = 0.7, col = "gray")
add.trajectories(TP, traj.pch = unique(as.numeric(traj.dat$fam) + 20), start.bg = 1, end.bg = 2)
legend("topright", levels(traj.dat$fam), pch =  unique(as.numeric(traj.dat$fam) + 20), bty="n")
#text(rownames(TA$pca$x), x=TP$pc.points[,"PC1"], y=TP$pc.points[,"PC2"], cex=0.2)


#make actual large bgPCA plot to highlight specific comparisons 

pdf(file=file.path("trajectory/Phyllo PC12.pdf"), width=7, height=7)

plot(gPCA_out$Scores[,1], pch=pchs, gPCA_out$Scores[,2], col=alpha(cols, 0.2), ylim=c(max(gPCA_out$Scores[,2])*1.5, min(gPCA_out$Scores[,2])), xlab=NA, ylab=NA, xaxt="n", yaxt="n") #ylim has max the min to reverse the y axis to make it similar to the true PCA plot. Also the lower value (technically the max) is *1/5 to give extra room at the bottom of the plot
mtext("Between Group Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Between Group Principal Component 2", side=2, line=0.2, cex=1.5)
title(main="Phyllodactylidae Toe Pad Class Comparisons", line = 0.8, cex.main=1.5)

#highlight the points this analysis is focused on
out<-gPCA_out$Scores[rownames(gPCA_out$Scores) %in% rownames(cats),1:2]
points(out, pch=pchs[rownames(gPCA_out$Scores) %in% rownames(cats)], col=alpha(cols[rownames(gPCA_out$Scores) %in% rownames(cats)], 0.6))
rm(out)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],1], gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],1][hpts], gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],2][hpts], col="red", lwd=1.5)


#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2][hpts], col="blue", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2][hpts], col="orange", lwd=1.5)


segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2]))


#mean points
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],2]), pch=21, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2]), pch=21, col="black", bg="orange", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2]), pch=21, col="black", bg="blue", cex=1.2)

#P values 
#magnitudes
mag<-as.expression(bquote(bold(Magnitudes)))

BvD<-as.expression(bquote(Padlass-Basal~vs~Padless-Distal*":"~.(pvals_display(Mags["basal-family:distal-family"]))))

legend("bottomleft", legend = c(mag, BvD), bty = "n", cex=0.7, inset = c(-0.01,0))



#angles 
mag<-as.expression(bquote(bold(Angles)))

BvD<-as.expression(bquote(Padlass-Basal~vs~Padless-Distal*":"~.(pvals_display(Angs["basal-family:distal-family"]))))


legend("bottomright", legend = c(mag, BvD), bty = "n", cex=0.7)


dev.off()



rm(NewCoords)
rm(cats)
rm(TA)
rm(TP)
rm(traj.dat)
rm(BvD)
rm(hpts)
rm(mag)
rm(Mags)
rm(Angs)




#Within Diplo comparisons
cats<-catassign[catassign$Family=="Diplo",]
Coords<-dat.super$coords[,,catassign$Family=="Diplo"]


#have to duplicate padless specimens so that multiple comparisons can be made
#basal will pair with padless under the  basal-family
#incipient will pair with padless2 under the inc-family
#distal will pair with padless3 under the distal-family
#we'll compare padless vs padded within each of our mock families


toadd<-cats[which(cats$category=="padless"),]
toadd$Family<-"inc-family"
toadd$category<-"padless2"
cats[cats$category=="padless","Family"]<-"basal-family"
cats<-rbind(cats, toadd)

toadd$Family<-"distal-family"
toadd$category<-"padless3"
cats<-rbind(cats, toadd)

cats[cats$category=="incipient_pad","Family"]<-"inc-family"
cats[cats$category=="distal_pad","Family"]<-"distal-family"
cats[cats$category=="basal_pad","Family"]<-"basal-family"


toadd<-cats$category
toadd[toadd=="padless2"]<-"padless"
toadd[toadd=="padless3"]<-"padless"
toadd[toadd=="basal_pad"]<-"padded"
toadd[toadd=="distal_pad"]<-"padded"
toadd[toadd=="incipient_pad"]<-"padded"

cats<-cbind(cats, toadd)
colnames(cats)[4]<-"general_padcat"

#unique(paste(cats$general_padcat, cats$Family))

toadd<-Coords[,,which(cats$category=="padless")]

NewCoords<-array(data = NA, dim = c(dim(Coords)[1], dim(Coords)[2],  dim(Coords)[3]+dim(toadd)[3]+dim(toadd)[3] ))
NewCoords[,,1:dim(Coords)[3]]<-Coords
NewCoords[,,(dim(Coords)[3]+1):(dim(Coords)[3]+dim(toadd)[3])]<-toadd
NewCoords[,,(dim(Coords)[3]+dim(toadd)[3]+1):(dim(Coords)[3]+dim(toadd)[3]+dim(toadd)[3])]<-toadd
dimnames(NewCoords)[[3]]<-rownames(cats)


rownames(cats)==dimnames(NewCoords)[[3]]
#should all be true

rm(toadd)
rm(Coords)

#make a new object (it will be a "list") that contains our coord data, our clade assignments, and within group assignments ie pad class
#made composite list that includes coords, clade, and padclass
traj.dat<-list(coords=two.d.array(NewCoords), pad=as.factor(cats$general_padcat), fam=as.factor(cats$Family))

fit <- lm.rrpp(coords ~ fam * pad, data = traj.dat, iter = 10000) #should be 10000?
#reveal.model.designs(fit)
TA <- trajectory.analysis(fit, groups = as.factor(cats$Family), traj.pts = as.factor(cats$general_padcat), print.progress = TRUE)

rm(fit)




# Magnitude difference (absolute difference between path distances)
summary(TA, attribute = "MD") 
#Observed path distances by group

#basal-family distal-family    inc-family 
#0.7662249     0.7286835     0.3078917 

#Pairwise absolute differences in path distances, plus statistics
#                                    d UCL (95%)          Z     Pr > d
#basal-family:distal-family 0.0375414 0.2313131 -0.7264473 0.75472453
#basal-family:inc-family    0.4583332 0.2422440  2.9301841 0.00009999 *
#distal-family:inc-family   0.4207918 0.2535533  2.6307909 0.00049995 *


#padless to incipient is different than all the other padless vs padded comparisons
Mags<-summary(TA, attribute = "MD")$summary.table[,"Pr > d"]
names(Mags)<-rownames(summary(TA, attribute = "MD")$summary.table)
#Adjust the Pvals as needed
Mags<-p.adjust(p =  Mags, method = "fdr", n = 50)
Mags
#basal-family:distal-family    basal-family:inc-family   distal-family:inc-family 
#1.00000000                 0.00499950                 0.01249875 



# Correlations (angles) between trajectories
summary(TA, attribute = "TC", angle.type = "deg") 

#Pairwise correlations between trajectories, plus statistics
#                                   r    angle UCL (95%)        Z Pr > angle
#basal-family:distal-family 0.8213414 34.78070  20.52260 3.327462 0.00009999
#basal-family:inc-family    0.8435699 32.48096  21.52006 2.894526 0.00059994
#distal-family:inc-family   0.8248211 34.42964  22.44149 2.925190 0.00079992



#all different angles 
Angs<-summary(TA, attribute = "TC", angle.type = "deg")$summary.table[,"Pr > angle"]
names(Angs)<-rownames(summary(TA, attribute = "TC", angle.type = "deg")$summary.table)
#Adjust the Pvals as needed
Angs<-p.adjust(p =  Angs, method = "fdr", n = 50)
Angs
#basal-family:distal-family    basal-family:inc-family   distal-family:inc-family 
#0.0049995                  0.0133320                  0.0133320 



# Plot results just for confirmation. Will actually illustrate results on larger bgPCA plot with all the data on it
TP <- plot(TA, pch = as.numeric(traj.dat$fam) + 20, bg = as.numeric(traj.dat$pad), cex = 0.7, col = "gray")
add.trajectories(TP, traj.pch = unique(as.numeric(traj.dat$fam) + 20), start.bg = 1, end.bg = 2)
legend("topright", levels(traj.dat$fam), pch =  unique(as.numeric(traj.dat$fam) + 20), bty="n")
#text(rownames(TA$pca$x), x=TP$pc.points[,"PC1"], y=TP$pc.points[,"PC2"], cex=0.2)


#make actual large bgPCA plot to highlight specific comparisons 

pdf(file=file.path("trajectory/Diplo PC12.pdf"), width=7, height=7)

plot(gPCA_out$Scores[,1], pch=pchs, gPCA_out$Scores[,2], col=alpha(cols, 0.2), ylim=c(max(gPCA_out$Scores[,2])*1.5, min(gPCA_out$Scores[,2])), xlab=NA, ylab=NA, xaxt="n", yaxt="n") #ylim has max the min to reverse the y axis to make it similar to the true PCA plot. Also the lower value (technically the max) is *1/5 to give extra room at the bottom of the plot
mtext("Between Group Principal Component 1", side=1, line=0.3, cex=1.5)
mtext("Between Group Principal Component 2", side=2, line=0.2, cex=1.5)
title(main="Diplodactylidae Toe Pad Class Comparisons", line = 0.8, cex.main=1.5)

#highlight the points this analysis is focused on
out<-gPCA_out$Scores[rownames(gPCA_out$Scores) %in% rownames(cats),1:2]
points(out, pch=pchs[rownames(gPCA_out$Scores) %in% rownames(cats)], col=alpha(cols[rownames(gPCA_out$Scores) %in% rownames(cats)], 0.6))
rm(out)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],1], gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],1][hpts], gPCA_out$Scores[cats$species[cats$general_padcat=="padless"],2][hpts], col="red", lwd=1.5)


#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2][hpts], col="blue", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2][hpts], col="orange", lwd=1.5)

#plot convex hull around each focal category
hpts<-chull(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2])
hpts<-c(hpts, hpts[1]) 
lines(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1][hpts], gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2][hpts], col="green", lwd=1.5)



segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2]))
segments(x0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padless"],1]), y0 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padless"],2]), x1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1]), y1 = mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2]))




#mean points
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padless"],2]), pch=24, col="black", bg="red", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="inc-family padded"],2]), pch=24, col="black", bg="green", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="distal-family padded"],2]), pch=24, col="black", bg="orange", cex=1.2)
points(mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],1]), mean(gPCA_out$Scores[cats$species[paste(cats$Family, cats$general_padcat)=="basal-family padded"],2]), pch=24, col="black", bg="blue", cex=1.2)

#P values 
#magnitudes
mag<-as.expression(bquote(bold(Magnitudes)))

BvD<-as.expression(bquote(Padlass-Basal~vs~Padless-Distal*":"~.(pvals_display(Mags["basal-family:distal-family"]))))
BvI<-as.expression(bquote(Padlass-Basal~vs~Padless-Incipient*":"~.(pvals_display(Mags["basal-family:inc-family"]))))
DvI<-as.expression(bquote(Padlass-Distal~vs~Padless-Incipient*":"~.(pvals_display(Mags["distal-family:inc-family"]))))

legend("bottomleft", legend = c(mag, BvD, BvI, DvI), bty = "n", cex=0.7, inset = c(-0.01,0))



#angles 
mag<-as.expression(bquote(bold(Angles)))

BvD<-as.expression(bquote(Padlass-Basal~vs~Padless-Distal*":"~.(pvals_display(Angs["basal-family:distal-family"]))))
BvI<-as.expression(bquote(Padlass-Basal~vs~Padless-Incipient*":"~.(pvals_display(Angs["basal-family:inc-family"]))))
DvI<-as.expression(bquote(Padlass-Distal~vs~Padless-Incipient*":"~.(pvals_display(Angs["distal-family:inc-family"]))))


legend("bottomright", legend = c(mag, BvD, BvI, DvI), bty = "n", cex=0.7)

dev.off()



rm(NewCoords)
rm(cats)
rm(TA)
rm(TP)
rm(traj.dat)
rm(BvI)
rm(BvD)
rm(DvI)
rm(hpts)
rm(mag)
rm(Mags)
rm(Angs)



rm(cols)
rm(gPCA_out)
rm(pchs)


###### Node Height Test ####

### CHANGE THIS ###
#Change this to the location that you saved the "PIC shape.R" file
source("/Users/travishagey/Desktop/Hagey/Science/Projects/R codes/GM plots/PIC Shape Node Height Test/PIC shape.R")


#remove Sphenodon_punctatus from node height test
Gecko_tree.NH<-drop.tip(Gecko_tree, tip = "Sphenodon_punctatus")
Species_Mshape.NH<-Species_Mshape[,,-which(dimnames(Species_Mshape)[[3]]=="Sphenodon_punctatus")]

NH.test_out<-PIC.shape(phy = Gecko_tree.NH, dat.shape = Species_Mshape.NH)

lm.out<-lm(formula = log(Std.IC)~NodeHeights, data = NH.test_out)
summary(lm.out)
#Coefficients:
#                Estimate Std. Error t value Pr(>|t|)    
#   (Intercept) -3.258377   0.139819 -23.304   <2e-16 ***
#   NodeHeights -0.005638   0.002633  -2.141    0.035 *  

#NHtest plot colors 
pal<-colorRamp(c("black", "lavenderblush"))
NHcolors<-log(NH.test_out$Std.IC)-min(log(NH.test_out$Std.IC)) #subtract the lowest value to shift the lowest value to be 0
NHcolors<-NHcolors/max(NHcolors) #divide by the largest value to adjust the largest value to be 1
NHcolorsnames<-NHcolors

for(r in 1:nrow(pal(NHcolors))){
  #r<-1
  NHcolorsnames[r]<-rgb(red=pal(NHcolors)[r,1], green=pal(NHcolors)[r,2], blue=pal(NHcolors)[r,3], maxColorValue = 255)
}

tree.max<-max(nodeHeights(Gecko_tree.NH))

#make NH plots
pdf(file=file.path("NH Test/Node Height Test.pdf"), width = 7.55, height = 4)
par(mar=c(5, 4, 4, 6))

plot(NH.test_out$NodeHeights, log(NH.test_out$Std.IC), xlab=NA, ylab=NA, pch=21, bg=NHcolorsnames, yaxt="n", xlim=c(0, tree.max), xaxt="n")
axis(side=1, at = seq(0, tree.max, by=20), labels = c("0", "20", "40", "60", "80", paste(round(tree.max, 2))))

segments(x0=NH.test_out$NodeHeights, y0=log(NH.test_out$Std.IC), x1=NH.test_out$NodeHeights, y1=abs(max(log(NH.test_out$Std.IC))*2), col="lightgrey")
abline(v=tree.max, col="lightgrey", lty=2) #current day
mtext(side=1, line=2, text = "Current Day", at = tree.max)

points(NH.test_out$NodeHeights, log(NH.test_out$Std.IC), pch=21, bg=NHcolorsnames)


legend_image <- as.raster(matrix(colorRampPalette(colors = c("lavenderblush", "black"))(10), ncol=1))
#rasterImage(legend_image, xleft=par('usr')[1]-2, xright=par('usr')[1], xpd=T, ytop=max(NH.test_out$Std.IC), ybottom=min(NH.test_out$Std.IC))
rasterImage(legend_image, xleft=par('usr')[1]-2, xright=par('usr')[1], xpd=T, ytop=par('usr')[4], ybottom=par('usr')[3])

#min(log(NH.test_out$Std.IC))
#max(log(NH.test_out$Std.IC))
axis(2, at=seq(-5, -2, by=1), pos=par('usr')[1]-2, padj = 0.9, lwd.ticks=0.5, tck=-0.03)

abline(lm.out)

#calculate confidence intervals around linear model
newx = seq(min(NH.test_out$NodeHeights),max(NH.test_out$NodeHeights),by = 1)
conf_interval <- predict(lm.out, newdata=data.frame(NodeHeights=newx), interval="confidence", level = 0.95)

lines(newx, conf_interval[,2], lwd=1, lty=2)
lines(newx, conf_interval[,3], lwd=1, lty=2)

mtext(text="Node Height Above Root (Millions of Years)", side = 1, line = 2)
mtext(text="Natural Log Standardized Phylogenetic\nIndependent Contrasts of Toe Pad Shape", side = 2, line = 2)

#add stats values to plot
slope<-signif(lm.out$coefficients["NodeHeights"], digits = 2)
tval<-round(summary(lm.out)$coefficients["NodeHeights","t value"], digits = 2)
pval<-round(summary(lm.out)$coefficients["NodeHeights","Pr(>|t|)"], digits = 3)

mtext(side=1, line=-2.6, adj = 0.02, text = paste("slope = ", slope, sep=""), cex=0.75)
mtext(side=1, line=-1.8, adj = 0.02, text = paste("t value = ", tval, sep=""), cex=0.75)
mtext(side=1, line=-1.0, adj = 0.02, text = paste("slope p value = ", pval, sep=""), cex=0.75, font=2)


dev.off()

#make a list of data (PIC values) to color each internal branch with
#NH.test node output is in a different order that phytools uses, so we need to make a list in the correct order
#nodeHeights(phy) gets the node heights for the ancestor and derived nodes for each branch in the phy$edges order
#going to color each pair of condensed branches by the evolutionary rate estimated by their shared ancestral node (its PIC value)

branchcolors<-rep(NA, Nedge(Gecko_tree.NH))

for(b in 1:Nedge(Gecko_tree.NH)){
  #b<-1
  branchcolors[b]<-log(NH.test_out[which(round(NH.test_out$NodeHeights, 6)==round(nodeHeights(Gecko_tree.NH)[,1], 6)[b]), "Std.IC"])
}


#make phylogeny to go along with NH test plot
pdf(file=file.path("NH Test/Gecko Phylogeny.pdf"), width = 7.5, height = 6)

#plot(Gecko_tree, cex=0.5, no.margin=T)
plotBranchbyTrait(Gecko_tree.NH, x=1, mode="edges", prompt=FALSE, cex=0.3, no.margin=T, legend = F, edge.width=4, palette = colorRampPalette(c("black", "black")), tip.color="white" )   

#get node coordinates
last.plot <- get("last_plot.phylo", envir = ape::.PlotPhyloEnv)

# save the x,y coordinates of everything
x <- last.plot$xx
y <- last.plot$yy

# make sure things map to the default labelling scheme from plot.phylo
labels <- c(Gecko_tree.NH$tip.label,(length(Gecko_tree.NH$tip.label)+1):(length(Gecko_tree.NH$tip.label)+Gecko_tree.NH$Nnode))

nodes<-data.frame(row.names=labels, x=x, y=y)
nodes<-nodes[-c(1:Ntip(Gecko_tree.NH)),]

segments(x0 = nodes$x, y0=nodes$y, x1=nodes$x, y1=-max(nodes$y), col="lightgrey")
segments(x0 = tree.max, y0=min(nodes$y), x1=tree.max, y1=-max(nodes$y), col="lightgrey", lty=2) #current day


par(new=TRUE)
plotBranchbyTrait(Gecko_tree.NH, x=1, mode="edges", prompt=FALSE, cex=0.3, no.margin=T, legend = F, edge.width=4, palette = colorRampPalette(c("black", "black")), tip.color="white" )   
par(new=TRUE)

#update african Cnemaspis to new genus Ancylodactylus
Gecko_tree.NH$tip.label[Gecko_tree.NH$tip.label=="Cnemaspis_africana"]<-"Ancylodactylus_africanus"
Gecko_tree.NH$tip.label[Gecko_tree.NH$tip.label=="Cnemaspis_dickersonae"]<-"Ancylodactylus_dickersonae"
Gecko_tree.NH$tip.label[Gecko_tree.NH$tip.label=="Cnemaspis_uzungwae"]<-"Ancylodactylus_uzungwae"

plotBranchbyTrait(Gecko_tree.NH, x=branchcolors, mode="edges", prompt=FALSE, cex=0.3, no.margin=T, legend = F, edge.width=3, palette = colorRampPalette(c("black", "lavenderblush")) )   

#edgelabels(text=round(branchcolors, 4), frame = "none", cex=0.5)


dev.off()

rm(conf_interval)
rm(Gecko_tree.NH)
rm(last.plot)
rm(lm.out)
rm(NH.test_out)
rm(nodes)
rm(b)
rm(branchcolors)
rm(labels)
rm(legend_image)
rm(newx)
rm(NHcolors)
rm(NHcolorsnames)
rm(pval)
rm(r)
rm(slope)
rm(Species_Mshape.NH)
rm(tval)
rm(x)
rm(y)
rm(new.branch.length)
rm(pal)
rm(PIC.shape)
rm(std.cont)
rm(wt.mshape)
rm(tree.max)

######
###### Included Functions ####
###### FUNCTION GM_curve_1st_last_remove() - removes 1st and last semilandmark points ####

## Arguments
# TPS_location - the file directory location and file name of the TPS file to be used, taking into account the working directory 

GM_curve_1st_last_remove<-function(TPS_location){
  
  #To remove the first and last semilandmarks from each indivudals curves we need to read the TPS file as text, foregoing the use of geomorph's import function
  dat<-readLines(TPS_location)

  #We need to know number of landmarks, the number of curves, and  number of points in each curve
  #We can extract this data from the dat object
  #We assume all individuals have the same number of landmarks and same number of semilandmarks in their curves

  num_LM <- grep("LM=",dat) #Which lines have the text "LM" in them
  num_LM<-dat[num_LM[1]] #Get the first line with the text "LM" in it
  num_LM<-as.numeric(strsplit(num_LM, "=")[[1]][2]) #Split that line at the "=" and keep the text after the "="

  num_curves <- grep("CURVES=",dat) #Which lines have the text "CURVES" in them
  num_curves<-dat[num_curves[1]] #Get the first line with the text "CURVES" in it
  num_curves<-as.numeric(strsplit(num_curves, "=")[[1]][2]) #Split that line at the "=" and keep the text after the "="

  #Now that we know the number of curves, we need the number of points per curve 
  num_points<-matrix(ncol=1, nrow=num_curves) #make a blank matrix with a row for each curve

  for(c in 1:num_curves){ #Run a loop for each curve
    #c<-1
    out<-dat[grep("POINTS=",dat)[c]] #get the "POINTS = " line for the c-th curve
    num_points[c,1]<-as.numeric(strsplit(out, "=")[[1]][2]) #Split that line at the "=" and keep the text after the "="
  }

  #Now we need to figure out which landmarks to remove
  #for each curve, figure out the landmark numbers, which is the number of landmarks plus previous curves

  to_remove<-matrix(ncol=0, nrow=0) #make a blank matrix which will be populatioed by the semilandmarks we need to drop

  for(c in 1:num_curves){ #Run a loop for each curve
    #c<-1
    pts<-1:num_points[c,] #get a set of numbers for the c-th curve 1 through the number of semilandmarks in that curve
    pts<-pts+num_LM #add the number of landmarks
    pts<-pts+sum(num_points[0:(c-1),]) #add the number of semilandmarks in the previous curves
    out<-c(pts[1], pts[length(pts)]) #get the first and last semilandmark numbers of our focal curve
    to_remove<-c(to_remove, out) #add those 2 semilandmark numbers to our "to_remove" matrix
  }

  to_remove

}



###### FUNCTION GM_IndvPlot () - plot/visualize landmarks and curves to assess order #####

## Arguments

# gpagen_output - This function uses geomorphs gpagen function output (aligned data)
# TPS_raw - It also unaligned data imported via geomorph (after first and last semilandmarks have been removed)
# slider_dat - It also uses the same landmark slider file
# plot_destination - The user need to dictate a save location for the created plots to go 

GM_IndvPlot<-function(gpagen_output, TPS_raw, slider_dat, plot_destination){
  
  library(shape) #This function requires the "Arrows" function from the shape library
  #gpagen_output<-dat.super
  #TPS_raw<-dat.raw
  #slider_dat<- sliders
  #plot_destination<-"Landmark Diagnostics/"
  
  
  curves<-as.list(as.data.frame(t(slider_dat))) #use the slider file to get the semilandmark relationships for each curve 
  
  #we need to build a set of landmarks and semilandmarks for each curve
  for(r in 1:length(curves)){ #Run a loop for each curve
    #r<-2
    second_tolast<-lapply(curves, function(x) x[length(x) -1]) #get each semilandmark from the slider file (ie the 2nd to last elements of each object in the list)
    
    if(curves[[r]][1] %in% second_tolast){       #for each semilandmark, if the point before it is also a semilandmark... 
      togo<-which(second_tolast==curves[[r]][1]) #Find the preceding semilandmark in the slider file
      curves[[togo]]<-c(curves[[togo]], curves[[r]][3]) #ammend the landmark (or semilandmark) that follows the current landmark to the preceding semilandmark's slider line
      curves[[r]]<-NA #clear the current semilandmarks line in the slider file
    }
  }
  
  curves<-curves[!is.na(curves)] #remove the empty lines from the slider file
  num_curves<-length(curves) #get the number of curves
  names(curves)<-paste("semi", 1:num_curves, sep="") #rename the objects sequentially
  #This should result in a list of objects, one for each curve, that lists the landmarks and semilandmarks in that curve in the correct order
  
  
  hard<-c(paste(lapply(curves, function(x) x[1])), paste(lapply(curves, function(x) x[length(x)]))) #get the first and last landmark from each curve
  hard<-sort(unique(as.numeric(hard))) #transform them into numeric variables and sort them
  
  #start translating variables 
  gpagen_output<-gpagen_output$coords #extract just the coordinate data from the gpagen output data
  nsamples<-dim(TPS_raw)[3] #get the number of individuals that we'll need plots for
  
  for (spec_num in 1:nsamples) {  #start cycling through individuals
    #spec_num<-199
    
    #Get that individual's data
    specimen_aligned<-gpagen_output[,,spec_num]
    specimen_raw<-TPS_raw[,,spec_num]
    
    Indv<-dimnames(gpagen_output)[[3]][spec_num] #get the individuals name 
    
    name1<-Indv
    #These sample names cannot contain pathway information. They must be cropped to only include the specimen name
    #geomorph automatically removes the filename extension from the specimen names 
    
 
    #set raw all positive, near 0
    specimen_raw[,1]<-specimen_raw[,1]-(min(specimen_raw[,1])) #shift x values
    specimen_raw[,2]<-specimen_raw[,2]-(min(specimen_raw[,2])) #shift y values
  
    
    #start making plots
    pdf(file=(file.path(paste(plot_destination, name1, ".pdf", sep=""))), width=7, height = 4)
    
    par(mfrow=c(1,2), mar=c(1,3,1,1), pty="s") 
    
    
    #unaligned points first 
    plot(specimen_raw[,1], specimen_raw[,2], xlim=c(min(specimen_raw), max(specimen_raw)), ylim=c(min(specimen_raw), max(specimen_raw)), xlab=NA, ylab=NA, col="white", main="Unaligned",  cex.axis=0.5)
    #mtext("(mm)", side=1, line=2) #this assumes your scale in TPSDig was in millimeters
    #mtext("(mm)", side=2, line=2) #this assumes your scale in TPSDig was in millimeters
    
    
    mtext(text=Indv, cex=0.7, side = 3, outer = T, line = -2)
    
    for (l in 1:num_curves) { #plot each curve
      #l<-1
      lines(x=c(specimen_raw[curves[[l]],1]), y=specimen_raw[curves[[l]],2], col=rainbow(num_curves)[l], type="o", pch=19, cex=0.3)
      Arrows(x0=specimen_raw[curves[[l]][length(curves[[l]])-1],1], x1=specimen_raw[curves[[l]][length(curves[[l]])],1], y0=specimen_raw[curves[[l]][length(curves[[l]])-1],2], y1=specimen_raw[curves[[l]][length(curves[[l]])],2], col=rainbow(num_curves)[l], arr.type="triangle", arr.length = 0.15, arr.width=0.15, segment=F)
    }
    #This function plots each curve, using a unique sequential color for each curve, ending with an arrow
    
    points(specimen_raw[hard,], pch=19, cex=0.5) #add landmarks
    text(specimen_raw[hard,], label=hard, cex=0.5, pos=3, offset=0.2) #adds  labels to the landmarks

    
    
    #aligned data
    plot(specimen_aligned[,1], specimen_aligned[,2], xlim=c(min(specimen_aligned), max(specimen_aligned)), ylim=c(min(specimen_aligned), max(specimen_aligned)), xlab=NA, col="white", main="Aligned",  cex.axis=0.5)
    
    for (l in 1:num_curves) { 
      #l<-1
      lines(x=c(specimen_aligned[curves[[l]],1]), y=specimen_aligned[curves[[l]],2], col=rainbow(num_curves)[l], type="o", pch=19, cex=0.3)
      Arrows(x0=specimen_aligned[curves[[l]][length(curves[[l]])-1],1], x1=specimen_aligned[curves[[l]][length(curves[[l]])],1], y0=specimen_aligned[curves[[l]][length(curves[[l]])-1],2], y1=specimen_aligned[curves[[l]][length(curves[[l]])],2], col=rainbow(num_curves)[l], arr.type="triangle", arr.length = 0.15, arr.width=0.15, segment=F)
    }
    #This function plots each curve, using a unique sequencial color for each curve, ending with an arrow
    
    points(specimen_aligned[hard,], pch=19, cex=0.5)
    text(specimen_aligned[hard,], label=hard, cex=0.5, pos=3, offset=0.2) #adds labels to the hard landmarks
    
    dev.off()
    
  }
  
}





###### FUNCTION GM_Indv_Landmark_Outliers () - detect and plot intra-landmark outliers #####

## Arguments
# gpagen_output - This function uses geomorph's gpagen function output (aligned data)
# plot_destination - The user needs to provide a save location for the created plots to go 

#Uses two approaches to determine if an individual's landmarks are significant different than all the other specimens 
#For each landmark, we calculate 
#The distance from the mean for each landmark and check to see if it is farther than 1.5 times the interquaretile distance from the mean (red circle)
#We calculate an eclipse 3 standard deviations from the mean in 2D landmark x,y space, and ask if any individual's landmark falls outside beyond that distance (blue eclipse)

GM_Indv_Landmark_Outliers<-function(gpagen_output, plot_destination){
  library(plotrix) #This function requires the "draw.circle" function from the shape plotrix
  library(phonTools) #this function calculates standard deviation ellipses around a group of observations
  library(sp) #check to see if points are within a polygon
  #gpagen_output<-dat.super
  #plot_destination<-"Indv Landmark Outliers/"
  
  #Will use two methods to determine if landmarks are significantly distant from the mean
  #IQR uses interquartile range to estimate landmarks significantly distant from the mean 
  #SD estimates landmarks 1.96 standard deviations from the mean
  
  #get just the coords part of the procrustes alignment output
  gpagen_output_full<-gpagen_output #keep the full output for later
  gpagen_output<-gpagen_output$coords
  
  #For each individual landmark we will investigate if there are any specimens with divergent landmarks
  for(land in 1:dim(gpagen_output)[1]){
    #land<-37
    
    #Need to extract this round's specific landmark from each specimen 
    out<-cbind(gpagen_output[land,1,], gpagen_output[land,2,])
    #determine mean landmark location 
    mean.land<-apply(out, MARGIN=2, FUN=mean)
    
    #determine each landmark's distance from the mean
    dists<-rep(NA, nrow(out)) #make a empty data set to receive distances 
    names(dists)<-rownames(out) #get specimen names from "out"
    for(l in 1:nrow(out)){ #for each specimen, calculate the distance if its landmark from the mean landmark 
      #l<-1
      dists[l]<-sqrt(((out[l,1]-mean.land[1])^2)+((out[l,2]-mean.land[2])^2))
    }
    dists<-sort(dists, decreasing = TRUE) #sort distances largest to smallest  
    
    
    #get quartile data
    med<-as.numeric(summary(dists)[3]) #median value
    #interquartile range (IQR); distance from 3rd to 1st quartile (3rd - 1st quartile)
    #IQR * 1.5
    #1st quartile - IQR = lower limit
    #3rd quartile + IQR = upper limit
    #Anything below lower limit or above upper limit would be considerd outliers
    LL<-as.numeric(summary(dists)[2] - (1.5 * (summary(dists)[5] - summary(dists)[2])))
    UL <- as.numeric(summary(dists)[5] + (1.5 * (summary(dists)[5] - summary(dists)[2])))
    
    #also estiamte SD's 
    SD_ellipse<-sdellipse (out, stdev = 3, lwd = 3, col = 2, show=FALSE)
    SD_out<-point.in.polygon(point.x = out[,1], point.y = out[,2], pol.x = SD_ellipse[,1], pol.y = SD_ellipse[,2])
    names(SD_out)<-rownames(out)
    #0's mean outside of polygon, 1's mean inside polygon
    
    
    #make plots
    pdf(file=(file.path(paste(plot_destination, "Landmark_", land, ".pdf", sep=""))), width=7, height = 7)
    par(mfrow=c(3,1), mar=c(1,3.5,1,1)) 
    
    #first plot shows landmark locations 
    plot(gpagen_output[,1,], gpagen_output[,2,], asp=1, col="grey", pch=19, cex=0.5, xlab=NA, ylab=NA, xaxt="n", yaxt="n", main=paste("Landmark", land))
    points(gpagen_output_full$consensus, col="dimgrey", pch=19, cex=0.5)
    
    points(out, col="black", pch=19, cex=0.5)
    
    legend("topleft", c("IQR", "3SD", "Both"), text.col = c("red", "blue", "green"), bty = "n")
    
    #IQR
    draw.circle(x=mean.land[1], y=mean.land[2], radius=UL, border="red")
    
    if (any(dists >= UL)) {
      points(out[names(which(dists >= UL)),1], out[names(which(dists >= UL)),2], pch = 19, col = "red", cex=0.5)
    }
    
    #SD ellipse
    points(SD_ellipse, type = "l", col="blue")
    
    if (any(SD_out==0)) {
      points(out[SD_out==0,1], out[SD_out==0,2], col="blue", cex=0.5, pch=19)
    }
    
    #outliers for both methods in green
    both<-intersect(names(which(dists >= UL)), names(which(SD_out==0)))
    
    if (length(both)>0){
      points(out[both,1], out[both,2], col="green", cex=0.5, pch=19)
    }
    
    
    #2nd plot shows landmark distances from mean in quartiles and SD's
    plot(dists, ylab = NA, pch = 19, xlab = NA, xaxt = "n")
    mtext(text = "Distance from Mean", side = 2, line = 2)
    
    abline(h = med, col = "black")
    abline(h = LL, lty = 2, col = "red")
    abline(h = UL, lty = 2, col = "red")
    
    text(x = length(dists), y = LL, labels = "lower quartile", col = "red", cex = 0.5, adj = c(0.5, 1))
    text(x = length(dists), y = med, labels = "median", col = "black", cex = 0.5, adj = c(0.5, -0.5))
    text(x = length(dists), y = UL, labels = "upper quartile", col = "red", cex = 0.5, adj = c(0.5, -0.5))
    
    if (any(dists >= UL)) {
      points(dists[which(dists >= UL)], pch = 19, col = "red")
      text(dists[which(dists >= UL)], labels = names(dists)[which(dists >= UL)], col = "red", adj = 0.8, pos = 4, cex = 0.5)
    }
    
    
    if (any(SD_out==0)) {
      toplot<-names(dists) %in% names(which(SD_out==0))
      points(y=dists[toplot], x=c(which(toplot)), col="blue", pch=19)
      text(y=dists[toplot], x=c(which(toplot)), labels = names(dists)[toplot], col = "blue", adj = 0.8, pos = 4, cex = 0.5)
    }
    
    
    if (length(both)>0){
      points(which(names(dists) %in% both), dists[both], col="green", pch=19)
      text(y=dists[both], x=which(names(dists) %in% both), labels = both, col = "green", adj = 0.8, pos = 4, cex = 0.5)
      
    }
    
    
    
    plot.new()
    #get names from both sets
    if (any(dists >= UL)  | (any(SD_out==0))){
      names<-unique(c(names(which(dists >= UL)), names(which(SD_out==0))))
      text(0,seq(from=1, to=0.1, length=length(names)), names, adj=0 )
    }
    
    dev.off()
    
  }
  
}



###### FUNCTION GM_shape_plot()  plots toepad shape and fills in color#####

##Arguments 
#toe_col toe background color
#lam_col color of lamellae 
#xcoords  x coordinates from data
#ycoords  y coordinates from data 

# FUNCTION
GM_shape_plot<-function(xcoords, ycoords, toe_col="gray", lam_col="dimgray"){
  outer<-c(1,12:24,3,25:37,5,38:50,4,51:63,2,1)
  polygon(xcoords[outer],ycoords[outer], border=NA,col=toe_col)
  
  lam1<-c(6,64:76,7,77:89,6)
  polygon(xcoords[lam1],ycoords[lam1], border=NA,col=lam_col)
  
  lam2<-c(8,90:102,9,103:115,8)
  polygon(xcoords[lam2], ycoords[lam2], border=NA,col=lam_col)
  
  lam3<-c(10,116:128,11,129:141,10)
  polygon(xcoords[lam3], ycoords[lam3], border=NA,col=lam_col)
}

######

